# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "04/03/2019"


from silx.utils.enum import Enum as _Enum

from ...utils import _assert_param_instance
from .anglemode import CorAngleMode
from tomwer.core.utils import getParametersFromParOrInfo
from .projectiontype import ProjectionType
from .mode import AxisMode
from collections import namedtuple
from silx.io.url import DataUrl
import silx.io.utils
import logging
import numpy
import os
import h5py
from typing import Union
from tomwer.core.process.task import Task
from silx.utils.deprecation import deprecated

try:
    from nabu.preproc.phase import PaganinPhaseRetrieval
except ImportError:
    from tomwer.third_party.nabu.preproc.phase import PaganinPhaseRetrieval

_logger = logging.getLogger(__name__)


_calculation_conf = namedtuple(
    "_calculation_conf", ["projection_type", "paganin"]
)  # noqa

_WITH_PAG = "withpag"
_NO_PAG = "nopag"


DEFAULT_CMP_THETA = 10
DEFAULT_CMP_SUBSAMPLING_Y = 10
DEFAULT_CMP_OVERSAMPLING = 4
DEFAULT_CMP_TAKE_LOG = True


class AxisCalculationInput(_Enum):
    """Define the different mode of input the user can have for axis calculation"""

    emission = _calculation_conf(ProjectionType.absorption, False)
    transmission = _calculation_conf(ProjectionType.transmission, False)
    transmission_pag = _calculation_conf(ProjectionType.transmission, True)

    def name(self):
        if self.value.paganin is True:
            return " ".join((self.value.projection_type.value, "paganin"))
        else:
            return self.value.projection_type.value

    def to_dict(self):
        pag_text = _WITH_PAG if self.value.paganin is True else _NO_PAG
        return "_".join((self.value.projection_type.value, pag_text))

    @classmethod
    def from_value(cls, value):
        if type(value) is str and len(value.split("_")) == 2:
            proj_type, pag = value.split("_")
            value_pag = True if pag == _WITH_PAG else False
            value_proj = ProjectionType.from_value(proj_type)
            return cls.from_value(_calculation_conf(value_proj, value_pag))
        elif isinstance(value, cls):
            return value
        else:
            for member in cls:
                if value in (member.value, member.name()):
                    return member
            raise ValueError("Cannot convert: %s" % value)


class AxisResource(object):
    """Helper for axis file relative stuff"""

    _PAGANIN_CONFIG = {"distance": 100e-3, "energy": 35, "delta_beta": 1e3}
    """Paganin configuration for axis calculation. To simplify we have only one
    static configuration for now. Otherwise complicate stuff"""

    def __init__(self, url):
        assert url is None or isinstance(url, DataUrl)
        assert url is None or url.is_valid()
        self.__url = url
        self.__raw_data = None
        self.__norme_data = None
        self.__norm_paganin = None

    def __str__(self):
        return f"{type(self)}, url: {self.__url.path() if self.__url else None}"

    @property
    def url(self):
        return self.__url

    @property
    def data(self):
        """

        :return: 2D numpy.array
        """
        if self.__url is None:
            return None
        if self.__raw_data is None:
            self.__raw_data = silx.io.utils.get_data(self.url)
        return self.__raw_data

    @data.setter
    def data(self, data):
        self.__raw_data = data

    @property
    def normalized_data(self):
        return self.__norme_data

    @normalized_data.setter
    def normalized_data(self, data):
        self.__norme_data = data

    def normalize_data(self, scan, log_):
        """Normalize data for axis calculation"""
        if self.__url is None:
            return None
        else:
            self.__norme_data = scan.data_flat_field_correction(data=self.data)
            if log_ is True:
                self.__norme_data[
                    numpy.isnan(self.__norme_data)
                ] = self.__norme_data.max()
                if self.__norme_data.max() < 0:
                    _logger.error("max data value < 0 unable to compute log")
                else:
                    try:
                        data = self.__norme_data
                        data[data <= 1] = 1
                        self.__norme_data = -numpy.log(data)
                    except Exception as e:
                        _logger.error("Fail to apply log on radio" + e)
            else:
                self.__norme_data[
                    numpy.isnan(self.__norme_data)
                ] = self.__norme_data.min()

        return self.__norme_data

    @property
    def normalized_data_paganin(self):
        """

        :return: data processed by the pagagnin phase retrieval
        :rtype: 2D numpy.array
        """
        if self.__url is None:
            return None
        if self.__norm_paganin is None:
            if self.normalized_data is None:
                raise ValueError("data should be normalized before applying " "paganin")
            else:
                data = self.normalized_data
                phase_retrieval = PaganinPhaseRetrieval(
                    data.shape, **AxisResource._PAGANIN_CONFIG
                )
                self.__norm_paganin = phase_retrieval.apply_filter(data)
        return self.__norm_paganin

    @normalized_data_paganin.setter
    def normalized_data_paganin(self, data):
        self.__norm_paganin = data

    def __eq__(self, other):
        if not isinstance(other, AxisResource):
            return False
        if self.url is None and other.url is None:
            return True
        elif self.url is None or other.url is None:
            return False
        else:
            return self.url == other.url


class AxisRP:
    """
    Configuration class for a tomwer :class:`AxisProcess`

    note: every modification on the parameters will process a call fo changed
    except `axis_url_1` and `axis_url_2` which will produce a call to the
    dedicated axis_url_changed
    """

    AXIS_POSITION_PAR_KEY = "ROTATION_AXIS_POSITION"
    """Key used for the axis position in par files"""

    def __init__(self):
        self.__mode = AxisMode.manual
        """Mode used for defining the COR (center of rotation)"""
        self.__relative_value = None
        """Value of the center of rotation in [0; image_width].
        None is not processing"""
        self.__absolute_value = None
        """Value of the center of rotation in [-image_width/2; image_width/2].
        None is not processing"""
        self.__angle_mode = CorAngleMode.use_0_180
        """Angle to use for radios"""
        self.__estimated_cor = 0.0
        """Position to use for near calculation"""
        self.__axis_url_1 = AxisResource(url=None)
        """first data url to use for axis cor calculation"""
        self.__axis_url_2 = AxisResource(url=None)
        """second data url to use for axis cor calculation"""
        self.__calculation_input_type = AxisCalculationInput.transmission
        """Type of input (emission, absorption, with or without paganin)"""
        self.__use_sinogram = False
        """Do we want to use the sinogram of the radios for computing center
        of rotation"""
        self.__sinogram_line = "middle"
        """Line of the radios to use for getting the sinogram"""
        self.__sinogram_subsampling = 10
        """if use sinogram activate, we can use a subsampling to reduce
        computation time"""
        self.__look_at_stdmax = False
        """do the near search at X position which as the max Y column standard
        deviation"""
        self.__near_wx = 5
        """do the near search in an X window of size +-near_wx"""
        self.__fine_stepx = 0.1
        """shift step x for fine shifting image"""
        self.__scale_img2_to_img1 = False
        """do image scaling"""
        self.__padding_mode = None
        self.__frame_width = None
        self.__side = "all"
        """side of the cor. Requested by nabu cor algorithms growing-window
        and sliding-window"""
        self.__flip_lr = True
        self.__composite_options = {
            "theta": DEFAULT_CMP_THETA,
            "subsampling_y": DEFAULT_CMP_SUBSAMPLING_Y,
            "oversampling": DEFAULT_CMP_OVERSAMPLING,
            "take_log": DEFAULT_CMP_TAKE_LOG,
        }
        """specific options for composite cor search"""

    @property
    def mode(self):
        return self.__mode

    @mode.setter
    def mode(self, mode):
        if isinstance(mode, str):
            try:
                name = mode
                if name in ("global_", "global"):
                    name = AxisMode.global_.name
                elif name == "accurate":
                    _logger.info(
                        "convert axis mode {} to {} (renamed)"
                        "".format(name, AxisMode.centered.name)
                    )
                    name = AxisMode.centered.name
                try:
                    _mode = getattr(AxisMode, name)
                except Exception:
                    _mode = AxisMode.from_value(name)
            except Exception:
                raise ValueError("Fail to create axis mode from %s" % mode)
        else:
            _assert_param_instance(mode, AxisMode)
            _mode = mode
        self.__mode = _mode
        self.changed()

    @property
    def frame_width(self) -> Union[None, float]:
        return self.__frame_width

    @frame_width.setter
    def frame_width(self, value):
        if not isinstance(value, (type(None), float, int)):
            raise TypeError(
                "Value is expected to be None or a float. Not {}".format(type(value))
            )
        if value is None:
            self.__frame_width = value
        else:
            self.__frame_width = float(value)

    @property
    def angle_mode(self):
        return self.__angle_mode

    @angle_mode.setter
    def angle_mode(self, mode):
        if type(mode) is str:
            mode = CorAngleMode.from_value(mode)

        if self.__angle_mode != mode:
            self.__angle_mode = mode
            self.changed()

    @property
    @deprecated(replacement="relative_cor_value", since_version="0.7")
    def value_ref_tomwer(self):
        return self.relative_cor_value

    @property
    def relative_cor_value(self):
        return self.__relative_value

    @property
    def absolute_cor_value(self):
        return self.__absolute_value

    def set_relative_value(self, value):
        _assert_param_instance(value, (int, float, str, type(None)))
        if value is None:
            changed = self.__relative_value is not None
        elif isinstance(value, str) and value == "...":
            changed = self.__relative_value != "..."
        else:
            changed = self.__relative_value != float(value)
        if changed:
            if value is None or (isinstance(value, str) and value == "..."):
                self.__relative_value = value
                self.__absolute_value = value
            else:
                self.__relative_value = float(value)
                if self.frame_width is not None:
                    self.__absolute_value = (
                        self.__relative_value + self.frame_width / 2.0
                    )
            self.changed()

    @deprecated(replacement="set_relative_value", since_version="0.7")
    def set_value_ref_tomwer(self, value):
        self.set_relative_value(value=value)

    @property
    @deprecated(replacement="absolute_cor_value", since_version="0.7")
    def value_ref_nabu(self):
        return self.__absolute_value

    @property
    def estimated_cor(self):
        return self.__estimated_cor

    @estimated_cor.setter
    def estimated_cor(self, value):
        if self.__estimated_cor != value:
            self.__estimated_cor = value
            self.changed()

    @property
    def side(self):
        return self.__side

    @side.setter
    def side(self, side):
        if side not in ("all", "left", "right", "center"):
            raise ValueError("side '{}' is not managed".format(side))
        if self.__side != side:
            self.__side = side
            self.changed()

    @property
    def axis_url_1(self):
        """the first file to be used for the axis calculation"""
        return self.__axis_url_1

    def _get_rsrc_frm_url(self, old_resource, new_url_):
        """
        Util function to compare new resource / url with an existing one.
        """
        new_resource = None
        changed = False
        if isinstance(new_url_, AxisResource):
            if old_resource != new_url_:
                changed = True
                new_resource = new_url_
        elif isinstance(new_url_, str):
            if new_url_ == "":
                new_resource = AxisResource(None)
            else:
                new_resource = AxisResource(DataUrl(path=new_url_))
            if old_resource != new_resource:
                changed = True
        else:
            assert isinstance(new_url_, DataUrl) or new_url_ is None
            new_resource = AxisResource(new_url_)
            if old_resource != new_resource:
                changed = True
        return new_resource, changed

    @axis_url_1.setter
    def axis_url_1(self, url_):
        new_resource, changed = self._get_rsrc_frm_url(
            old_resource=self.__axis_url_1, new_url_=url_
        )
        if changed is True:
            self.__axis_url_1 = new_resource
            self.axis_urls_changed()

    @property
    def axis_url_2(self):
        """the second file to be used for the axis calculation"""
        return self.__axis_url_2

    @axis_url_2.setter
    def axis_url_2(self, url_):
        new_resource, changed = self._get_rsrc_frm_url(
            old_resource=self.__axis_url_2, new_url_=url_
        )
        if changed is True:
            self.__axis_url_2 = new_resource
            self.axis_urls_changed()

    @property
    def flip_lr(self):
        return self.__flip_lr

    @flip_lr.setter
    def flip_lr(self, flip: bool):
        if not isinstance(flip, bool):
            raise TypeError("flip should be a boolean")
        self.__flip_lr = flip

    @property
    def projection_type(self):
        return self.__calculation_input_type.value.projection_type

    @property
    def paganin_preproc(self):
        return self.__calculation_input_type.value.paganin

    @property
    def calculation_input_type(self):
        return self.__calculation_input_type

    @calculation_input_type.setter
    def calculation_input_type(self, type_):
        assert isinstance(type_, AxisCalculationInput)
        value = AxisCalculationInput.from_value(type_)
        if value != self.__calculation_input_type:
            self.__calculation_input_type = type_
            self.changed()

    @property
    def use_sinogram(self):
        return self.__use_sinogram

    @use_sinogram.setter
    def use_sinogram(self, sinogram):
        if self.__use_sinogram != sinogram:
            self.__use_sinogram = sinogram
            self.changed()

    @property
    def sinogram_line(self):
        return self.__sinogram_line

    @sinogram_line.setter
    def sinogram_line(self, line):
        if line == "":
            if self.__sinogram_line is not None:
                self.__sinogram_line = None
                self.changed()
        else:
            if line == "middle":
                line = line
            elif line is not None:
                line = int(line)
            if self.__sinogram_line != line:
                self.__sinogram_line = line
                self.changed()

    @property
    def sinogram_subsampling(self) -> int:
        """Subsample radio to speed up processing of the sinogram generation"""
        return self.__sinogram_subsampling

    @sinogram_subsampling.setter
    def sinogram_subsampling(self, subsampling: int) -> None:
        subsampling = int(subsampling)
        if subsampling != self.__sinogram_subsampling:
            self.__sinogram_subsampling = subsampling
            self.changed()

    @property
    def look_at_stdmax(self):
        return self.__look_at_stdmax

    @look_at_stdmax.setter
    def look_at_stdmax(self, stdmax):
        self.__look_at_stdmax = stdmax

    @property
    def near_wx(self):
        return self.__near_wx

    @near_wx.setter
    def near_wx(self, width):
        if self.__near_wx != width:
            self.__near_wx = width
            self.changed()

    @property
    def fine_step_x(self):
        return self.__fine_stepx

    @fine_step_x.setter
    def fine_step_x(self, step_size):
        if self.__fine_stepx != step_size:
            self.__fine_stepx = step_size
            self.changed()

    @property
    def scale_img2_to_img1(self):
        return self.__scale_img2_to_img1

    @scale_img2_to_img1.setter
    def scale_img2_to_img1(self, scale):
        if self.__scale_img2_to_img1 != scale:
            self.__scale_img2_to_img1 = scale
            self.changed()

    @property
    def padding_mode(self):
        return self.__padding_mode

    @property
    def composite_options(self) -> dict:
        """return specific options for composite cor search"""
        return self.__composite_options

    @composite_options.setter
    def composite_options(self, opts: dict) -> None:
        """
        :param dict opts: options to use for the composite COR search
        :raises:
            * KeyError if some provided keys are not handled
            * TypeError if opts is not an instance of dictionary
        """
        if not isinstance(opts, dict):
            raise TypeError(
                f"opts is expected to be an instance of dict not {type(opts)}"
            )
        for key in opts.keys():
            if key not in ("theta", "oversampling", "subsampling_y", "take_log"):
                raise KeyError(f"{key} is not recogized")
        self.__composite_options = opts

    @padding_mode.setter
    def padding_mode(self, mode):
        if self.__padding_mode != mode:
            self.__padding_mode = mode
            self.changed()

    def changed(self):
        """callback to overwrite when the paramter value changed"""
        pass

    def set_position_frm_par_file(self, file_path, entry=None, force=False):
        """Set position value from a file

        :param str file_path: path to the par file
        :param bool force: if True, force the mode to read and set
                           `position_value` to None if failed to read the file.
                           Otherwise only `mode` and `position_value` when the
                           file read succeed.
        """

        def get_position_in_file():
            if not os.path.isfile(file_path):
                raise ValueError(file_path, "is not a valid file path")

            if h5py.is_hdf5(file_path):
                return self.get_cor_position_frm_tomwer_file(
                    file_path=file_path, entry=entry
                )
            elif file_path.lower().endswith(".par") or file_path.lower().endswith(
                ".info"
            ):
                _dict = getParametersFromParOrInfo(file_path)
                if self.AXIS_POSITION_PAR_KEY.lower() not in _dict:
                    err = "fail to find `%s` in %s" % (
                        self.AXIS_POSITION_PAR_KEY,
                        file_path,
                    )
                    _logger.info(err)
                    raise KeyError(err)
                else:
                    return _dict[self.AXIS_POSITION_PAR_KEY.lower()]
            else:
                with open(file_path, mode="r") as fd:
                    l1 = fd.readline()
                    values = l1.split(" ")
                    if len(values) > 0:
                        try:
                            res = float(values[0])
                        except Exception:
                            _logger.info(
                                "fail to load center of rotation " "from ", file_path
                            )
                            return None
                        else:
                            return res

        if force:
            try:
                position = get_position_in_file()
            except (KeyError, ValueError) as e:
                _logger.warning(e)
                position = None
        else:
            position = get_position_in_file()

        changed = self.mode != AxisMode.read or position != self.relative_cor_value
        self.mode = AxisMode.read
        if position is None:
            self.set_relative_value(None)
        else:
            self.set_relative_value(float(position))

        if changed:
            self.changed()

    @staticmethod
    def get_cor_position_frm_tomwer_file(
        file_path: str, entry=None
    ) -> Union[None, float]:
        """
        Parse the given .h5 file, looking for axis processes. If found some
        then return the first value discover

        :param str file_path: tomwer processes file (.h5)
        :return: center of rotation or none
        :rtype: Union[None, float]
        """
        from .axis import AxisProcess  # avoid cyclic import

        processes = Task.get_processes_frm_type(
            process_file=file_path, process_type=AxisProcess, entry=entry
        )
        if len(processes) == 0:
            _logger.warning("failed to find any information on center " "of rotation")
            return None
        elif len(processes) > 1:
            _logger.info("found more than one value computed for " "center of rotation")
        _logger.info(
            "load center of rotation from %s: %s"
            % (file_path, processes[0].results["center_of_rotation"])
        )
        cor = processes[0].results["center_of_rotation"]
        if cor in [None, ""]:
            return None
        else:
            return float(cor)

    def n_url(self):
        """

        :return: number of available url from url_1, url_2
        """
        n_url = 0
        if self.axis_url_1 and self.axis_url_1.url:
            n_url += 1
        if self.axis_url_2 and self.axis_url_2.url:
            n_url += 1
        return n_url

    def to_dict(self):
        # keep octave compatibility
        axis_urls_1 = self.axis_url_1.url
        if axis_urls_1 is None:
            axis_urls_1 = ""
        else:
            axis_urls_1 = axis_urls_1.path()
        axis_urls_2 = self.axis_url_2.url
        if axis_urls_2 is None:
            axis_urls_2 = ""
        else:
            axis_urls_2 = axis_urls_2.path()

        _dict = {
            "MODE": self.mode.value,
            "POSITION_VALUE": self.relative_cor_value,
            "CALC_INPUT_TYPE": self.calculation_input_type.to_dict(),
            "ANGLE_MODE": self.angle_mode.value,
            "USE_SINOGRAM": self.use_sinogram,
            "SINOGRAM_LINE": self.sinogram_line if self.use_sinogram else "",
            "SINOGRAM_SUBSAMPLING": self.sinogram_subsampling,
            "AXIS_URL_1": axis_urls_1,
            "AXIS_URL_2": axis_urls_2,
            "LOOK_AT_STDMAX": self.look_at_stdmax,
            "NEAR_WX": self.near_wx,
            "FINE_STEP_X": self.fine_step_x,
            "SCALE_IMG2_TO_IMG1": self.scale_img2_to_img1,
            "NEAR_POSITION": self.estimated_cor,
            "PADDING_MODE": self.padding_mode,
            "FLIP_LR": self.flip_lr,
            "COMPOSITE_OPTS": self.composite_options,
        }
        return _dict

    @staticmethod
    def from_dict(_dict):
        axis = AxisRP()
        axis.load_from_dict(_dict=_dict)
        return axis

    def load_from_dict(self, _dict):
        if "MODE" in _dict:
            self.mode = _dict["MODE"]
        if "POSITION_VALUE" in _dict:
            self.set_relative_value(_dict["POSITION_VALUE"])
        if "CALC_INPUT_TYPE" in _dict:
            self.calculation_input_type = AxisCalculationInput.from_value(
                _dict["CALC_INPUT_TYPE"]
            )
        if "USE_SINOGRAM" in _dict:
            self.use_sinogram = _dict["USE_SINOGRAM"]
        if "ANGLE_MODE" in _dict:
            self.angle_mode = CorAngleMode.from_value(_dict["ANGLE_MODE"])
        if "SINOGRAM_LINE" in _dict:
            self.sinogram_line = _dict["SINOGRAM_LINE"]
        if "AXIS_URL_1" in _dict:
            self.axis_url_1 = _dict["AXIS_URL_1"]
        if "AXIS_URL_2" in _dict:
            self.axis_url_2 = _dict["AXIS_URL_2"]
        if "LOOK_AT_STDMAX" in _dict:
            self.look_at_stdmax = _dict["LOOK_AT_STDMAX"]
        if "NEAR_WX" in _dict:
            self.near_wx = _dict["NEAR_WX"]
        if "FINE_STEP_X" in _dict:
            self.fine_step_x = _dict["FINE_STEP_X"]
        if "SCALE_IMG2_TO_IMG1" in _dict:
            self.scale_img2_to_img1 = _dict["SCALE_IMG2_TO_IMG1"]
        if "NEAR_POSITION" in _dict:
            self.estimated_cor = float(_dict["NEAR_POSITION"])
        if "SINOGRAM_SUBSAMPLING" in _dict:
            self.sinogram_subsampling = _dict["SINOGRAM_SUBSAMPLING"]
        if "PADDING_MODE" in _dict:
            self.padding_mode = _dict["PADDING_MODE"]
        if "FLIP_LR" in _dict:
            self.flip_lr = bool(_dict["FLIP_LR"])
        if "COMPOSITE_OPTS" in _dict:
            self.composite_options = _dict["COMPOSITE_OPTS"]

    def copy(self, axis_params, copy_axis_url=True, copy_flip_lr=True):
        assert isinstance(axis_params, AxisRP)
        self.mode = axis_params.mode
        self.frame_width = axis_params.frame_width
        self.set_relative_value(axis_params.relative_cor_value)
        self.calculation_input_type = axis_params.calculation_input_type
        self.use_sinogram = axis_params.use_sinogram
        self.angle_mode = axis_params.angle_mode
        self.sinogram_line = axis_params.sinogram_line
        self.sinogram_subsampling = axis_params.sinogram_subsampling
        self.look_at_stdmax = axis_params.look_at_stdmax
        self.near_wx = axis_params.near_wx
        self.fine_step_x = axis_params.fine_step_x
        self.scale_img2_to_img1 = axis_params.scale_img2_to_img1
        self.estimated_cor = axis_params.estimated_cor
        self.padding_mode = axis_params.padding_mode
        self.side = axis_params.side
        self.composite_options = axis_params.composite_options
        if copy_axis_url:
            self.axis_url_1 = axis_params.axis_url_1
            self.axis_url_2 = axis_params.axis_url_2
        if copy_flip_lr:
            self.flip_lr = axis_params.flip_lr

    def __str__(self):
        return str(self.to_dict())

    def axis_urls_changed(self):
        """Callback when the axis url change"""
        pass

    def get_simple_str(self):
        results = "{}".format(self.mode.value)
        if self.mode in (AxisMode.growing_window, AxisMode.sliding_window):
            extra_info = "side: {}, use sinogram: {}".format(
                self.side, self.use_sinogram
            )
            results = ", ".join((results, extra_info))
        return results

    def get_nabu_cor_options(self) -> str:
        """return cor option for nabu"""
        if self.mode in (AxisMode.sliding_window, AxisMode.growing_window):
            options = "side='{}'".format(self.side)
        else:
            options = None
        return options



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "24/01/2017"


import logging
import os
import shutil
import tempfile
import unittest
import gc
from silx.gui.utils.testutils import TestCaseQt
from silx.gui import qt
from orangecontrib.tomwer.widgets.reconstruction.FtseriesOW import FtseriesOW
from tomwer.core.process.reconstruction.ftseries.params.fastsetupdefineglobals import (
    FastSetupAll,
)
from tomwer.core.process.reconstruction.ftseries.params.paganin import PaganinMode
from tomwer.core.process.reconstruction.ftseries.params.beamgeo import BeamGeoType
from tomwer.core.process.reconstruction.ftseries.params.reconsparams import _ReconsParam
from tomwer.core.process.reconstruction.ftseries.params.ft import (
    VolSelMode,
    FixedSliceMode,
)
from tomwer.gui.qtapplicationmanager import QApplicationManager
from tomwer.synctools.ftseries import QReconsParams
from tomwer.core.utils import ftseriesutils
from tomwer.core.utils.scanutils import MockEDF


logging.disable(logging.INFO)

# Makes sure a QApplication exists
_qapp = QApplicationManager()


class TestFTSerieWidget(TestCaseQt):
    """Simple unit test to test the start/stop observation button action"""

    def setUp(self):
        super(TestFTSerieWidget, self).setUp()
        self.owftseriewidget = FtseriesOW(_connect_handler=False)
        self._reconsParams = self.owftseriewidget.recons_params

        self.ftseriewidget = self.owftseriewidget._ftserie

        self.tempdir = tempfile.mkdtemp()
        MockEDF.mockScan(self.tempdir, nRadio=5, nRecons=1, nPagRecons=0, dim=10)

        self.h5_fname = os.path.join(self.tempdir, "octave.h5")
        ftseriesutils.generateDefaultH5File(self.h5_fname)
        self.savePath = tempfile.mkstemp(suffix=".h5")[1]
        self._reconsParams.copy(FastSetupAll.DEFAULT_VALUES)
        self.owftseriewidget.setAttribute(qt.Qt.WA_DeleteOnClose)

    def tearDown(self):
        os.unlink(self.savePath)
        if os.path.isdir(self.tempdir):
            shutil.rmtree(self.tempdir)
        self.owftseriewidget.close()
        self.owftseriewidget = None
        self.qapp.processEvents()
        gc.collect()
        super(TestFTSerieWidget, self).tearDown()

    def testAngleOffset(self):
        self.ftseriewidget.updatePath(self.tempdir)
        self.assertTrue(
            self.ftseriewidget.getReconsParamSetEditor().isParamH5Managed(
                "FT", "ANGLE_OFFSET"
            )
        )
        _qapp.processEvents()
        ft = self.ftseriewidget.getReconsParamSetEditor().getStructs()
        self.assertTrue(ft["FT"]["FIXEDSLICE"] == "middle")
        self.assertTrue(float(ft["FT"]["ANGLE_OFFSET_VALUE"]) == 0.0)

        self.owftseriewidget.show()

        self.ftseriewidget.save(self.savePath)
        _qapp.processEvents()

        readed = FastSetupAll()
        readed.readAll(self.h5_fname, 3.8)
        self.assertTrue(readed.structures["FT"]["ANGLE_OFFSET"] == 0)

        self._reconsParams.ft["ANGLE_OFFSET_VALUE"] = 1.0
        self.assertTrue(self._reconsParams.ft["ANGLE_OFFSET"] == 1)
        self.ftseriewidget.save(self.savePath)
        _qapp.processEvents()
        readed = FastSetupAll()
        readed.readAll(self.savePath, 3.8)
        self.assertTrue(readed.structures["FT"]["ANGLE_OFFSET_VALUE"] == 1)
        self.assertTrue(readed.structures["FT"]["ANGLE_OFFSET"] == 1)


class TestControlsFTSeriesWidget(TestCaseQt):
    """Simple test of the control button for the FTserieWidget"""

    def setUp(self):
        super(TestControlsFTSeriesWidget, self).setUp()
        # create octave h5 with default
        self.path = tempfile.mkdtemp()
        self.scan = MockEDF.mockScan(
            self.path, nRadio=5, nRecons=1, nPagRecons=0, dim=10
        )
        # one with some modifications
        self.modifySettings = os.path.join(self.path, "modifySettings.h5")
        ft = FastSetupAll()
        assert ft.structures["PAGANIN"]["MODE"] != 2
        ft.structures["PAGANIN"]["MODE"] = 2
        ft.writeAll(self.modifySettings, 3.8)
        # simulate an acquisition (files at least)
        self.widget = FtseriesOW(_connect_handler=False)
        self.widget.setDryRun(True)

    def tearDown(self):
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        if os.path.isdir(self.path):
            shutil.rmtree(self.path)
        self.qapp.processEvents()
        gc.collect()
        super(TestControlsFTSeriesWidget, self).tearDown()

    def testShow(self):
        """make sure the h5file is created in the scan directory"""
        self.widget.reprocess(self.scan)
        _qapp.processEvents()
        self.assertTrue(self.path == self.widget.scan.path)

    def testSave(self):
        """Make sure that the save action is correct"""
        self.widget.updatePath(self.scan)
        _qapp.processEvents()
        self.assertTrue(self.path == self.widget.scan.path)

        tmpH5File = tempfile.mkstemp(suffix=".h5")[1]
        self.widget.save(tmpH5File, displayInfo=False)

        ftOri = FastSetupAll()
        originalStructures = ftOri.structures

        ftSaved = FastSetupAll()
        ftSaved.readAll(filn=self.modifySettings, targetted_octave_version=3.8)
        self.assertTrue(ftSaved.structures["PAGANIN"]["MODE"].value == 2)
        ftSaved.structures["PAGANIN"]["MODE"] = originalStructures["PAGANIN"]["MODE"]
        self.assertTrue(originalStructures.to_dict() == ftSaved.structures.to_dict())

    def testLoad(self):
        """Make sure the load action is correct"""
        self.widget.updatePath(self.scan)
        _qapp.processEvents()
        self.assertTrue(self.path == self.widget.scan.path)
        self.widget.show()
        self.widget.load_recons_params(self.modifySettings)

        tmpH5File = tempfile.mkstemp(suffix=".h5")[1]
        self.widget.save(tmpH5File)
        self.assertTrue(self.fileContainEqual(self.modifySettings, tmpH5File))
        os.unlink(tmpH5File)

    def fileContainEqual(self, f1, f2):
        ft1 = FastSetupAll()
        ft1.readAll(filn=f1, targetted_octave_version=3.8)

        ft2 = FastSetupAll()
        ft2.readAll(filn=f2, targetted_octave_version=3.8)

        return ft1.structures.to_dict() == ft2.structures.to_dict()


class TestExploreDatasetForH5(TestCaseQt):
    """Make sure that in the case the octave_FT_params.h5 exists in a directory
    this will be the one picked by the editor if the 'explore scan' option
    is activated
    """

    def setUp(self):
        super(TestExploreDatasetForH5, self).setUp()
        self._reconsParams = QReconsParams()

        self.pathWithH5 = tempfile.mkdtemp()
        self.pathWithoutH5 = tempfile.mkdtemp()
        for folder in (self.pathWithH5, self.pathWithoutH5):
            MockEDF.mockScan(folder, nRadio=5, nRecons=1, nPagRecons=0, dim=10)

        self.ft = FastSetupAll()
        assert self.ft.structures["PAGANIN"]["MODE"] != 2
        self.ft.structures["PAGANIN"]["MODE"] = 2

        self.modifySettings = os.path.join(self.pathWithH5, "octave_FT_params.h5")
        self.ft.writeAll(self.modifySettings, 3.8)

        self.owWidget = FtseriesOW(
            _connect_handler=False, recons_params=self._reconsParams
        )
        self.widget = self.owWidget._ftserie

        self._reconsParams.copy(self.ft.DEFAULT_VALUES)

    def tearDown(self):
        if os.path.isfile(self.modifySettings):
            os.remove(self.modifySettings)
        if os.path.isdir(self.pathWithH5):
            shutil.rmtree(self.pathWithH5)
        if os.path.isdir(self.pathWithoutH5):
            shutil.rmtree(self.pathWithoutH5)
        self.owWidget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.owWidget.close()
        self.owWidget = None
        self.qapp.processEvents()
        gc.collect()
        super(TestExploreDatasetForH5, self).tearDown()

    def testExploreInFolderWithH5(self):
        """Make sure the widget if try to explore the scan dir will load
        the reconstruction parameters from it
        """
        self.widget.setH5Exploration(True)
        self.widget.updatePath(self.pathWithH5)
        self.assertTrue(self.pathWithH5 == self.widget.scan.path)
        self.assertTrue(self._reconsParams.paganin.mode.value == 2)

    def testExploreInFolderWithoutH5(self):
        """Make sure the widget if try to explore the scan dir will load
        the reconstruction parameters from it
        """
        self.widget.setH5Exploration(False)
        self.widget.updatePath(self.pathWithoutH5)
        self.assertTrue(self.pathWithoutH5 == self.widget.scan.path)
        self.assertTrue(self._reconsParams.paganin.mode.value != 2)

    def testNoExploreInFolderWithoutH5(self):
        self.widget.setH5Exploration(False)
        self.widget.updatePath(self.pathWithH5)
        self.assertTrue(self.pathWithH5 == self.widget.scan.path)
        self.assertTrue(self._reconsParams.paganin.mode.value != 2)

    def testNoExploreInFolderWithH5(self):
        self.widget.setH5Exploration(False)
        self.widget.updatePath(self.pathWithoutH5)
        self.assertTrue(self.pathWithoutH5 == self.widget.scan.path)
        self.assertTrue(self._reconsParams.paganin.mode.value != 2)


class TestLoadSaveH5File(TestCaseQt):
    """Simple test that make sure we are able to manage correctly input h5 file
    with the following behavior :
        - if some H5 parameter are missing fron the default set of parameter
        then we are loading the default parameter
        - if the h5 file as some extra H5 parameter or structure then the
        editor should also store them.
    """

    def setUp(self):
        super(TestLoadSaveH5File, self).setUp()
        self.path = tempfile.mkdtemp()
        MockEDF.mockScan(self.path, nRadio=5, nRecons=1, nPagRecons=0, dim=10)
        self.ft = FastSetupAll()
        self.file = os.path.join(self.path, "octave_FT_params.h5")

    def tearFown(self):
        del self.ft
        del self.file
        if os.path.isdir(self.path):
            shutil.rmtree(self.path)
        super(TestLoadSaveH5File, self).tearDown()

    def loadAndSave(self):
        # save actual FastSetup status
        self.ft.writeAll(self.file, 3.8)

        # create FTSerie, load and save dataparamValue = self.getDefaultValue(param)
        owWidgetWithout = FtseriesOW(_connect_handler=False)
        widgetWithout = owWidgetWithout._ftserie
        owWidgetWithout.setAttribute(qt.Qt.WA_DeleteOnClose)

        widgetWithout.setH5Exploration(True)
        widgetWithout.updatePath(self.path)
        widgetWithout.save(self.file)
        owWidgetWithout.close()
        # then get back what FTSerie has saved
        loader = FastSetupAll()
        loader.readAll(self.file, targetted_octave_version=3.8)
        return loader.structures

    def testDefaultFileWithExtraParameters(self):
        self.ft.structures["TEST"] = {"PARAM1": 1, "PARAM2": "dsadsa"}
        self.ft.structures["FT"]["PARAMTATA"] = 56.3

        savedData = self.loadAndSave()
        self.assertTrue("PARAMTATA" in savedData["FT"].all_params)
        self.assertTrue(savedData["FT"]["PARAMTATA"] == 56.3)


class TestSync(TestCaseQt):
    """Make sure all the tab from the editor are correctly sync with
    :class:`ReconsParam` instance"""

    def setUp(self):
        super(TestSync, self).setUp()
        self._reconsParams = QReconsParams()
        self.owWidget = FtseriesOW(_connect_handler=False)
        self.owWidget.setReconsParams(self._reconsParams)
        self.widget = self.owWidget._ftserie
        assert id(self.widget._recons_params) == id(self._reconsParams)
        assert id(self.widget._recons_params.pyhst) == id(self._reconsParams.pyhst)
        assert id(self.widget.getReconsParamSetEditor().reconsparams) == id(
            self._reconsParams
        )
        assert id(self.widget.getReconsParamSetEditor().reconsparams.pyhst) == id(
            self._reconsParams.pyhst
        )

    def tearDown(self):
        self.owWidget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.owWidget.close()
        self.owWidget = None
        super(TestSync, self).tearDown()

    def testFTRead(self):
        """Make sure modifications from the ReconsParam are take into account"""
        ftWidget = self.widget.getReconsParamSetEditor()._mainWidget
        assert self._reconsParams["FT"]["DO_TEST_SLICE"] == 1
        self._reconsParams["FT"]["DO_TEST_SLICE"] = 0
        self.assertFalse(ftWidget._qcbDoTestSlice.isChecked())

        assert self._reconsParams.to_dict()["FT"]["FIXEDSLICE"] == "middle"
        assert self._reconsParams["FT"]["FIXEDSLICE"] == FixedSliceMode.middle
        self._reconsParams["FT"]["FIXEDSLICE"] = FixedSliceMode.row_n
        self.assertTrue(ftWidget._qcbSelectSlice.currentText() == "row n")

        assert self._reconsParams["FT"]["CORRECT_SPIKES_THRESHOLD"] == 0.04
        self._reconsParams["FT"]["CORRECT_SPIKES_THRESHOLD"] = 1.23
        self.assertTrue(float(ftWidget._qleThresholdSpikesRm.text()) == 1.23)

        assert self._reconsParams["FT"]["RINGSCORRECTION"] == 0
        self._reconsParams["FT"]["RINGSCORRECTION"] = 1
        self.assertTrue(ftWidget._qcbRingsCorrection.isChecked())

        assert self._reconsParams["FT"]["VOLSELECTION_REMEMBER"] == 0
        self._reconsParams["FT"]["VOLSELECTION_REMEMBER"] = 1
        self.assertTrue(ftWidget._qcbVolSelRemenber.isChecked())

        assert self._reconsParams.to_dict()["FT"]["VOLSELECT"] == "total"
        assert self._reconsParams["FT"]["VOLSELECT"] == VolSelMode.total
        self._reconsParams["FT"]["VOLSELECT"] = VolSelMode.graphics
        self.assertTrue(ftWidget._qcbVolumeSelection.currentText() == "graphics")

    def testFTWrite(self):
        """Make sure user modification from the gui are updating the
        :class:`ReconsParam` instance"""
        ftWidget = self.widget.getReconsParamSetEditor()._mainWidget

        assert self._reconsParams["FT"]["DO_TEST_SLICE"] == 1
        ftWidget._qcbDoTestSlice.setChecked(False)
        self.qapp.processEvents()
        self.assertTrue(self._reconsParams["FT"]["DO_TEST_SLICE"] == 0)
        assert self._reconsParams["FT"]["FIXEDSLICE"] == FixedSliceMode.middle
        assert self._reconsParams.to_dict()["FT"]["FIXEDSLICE"] == "middle"

        ftWidget._qleSliceN.setText("4")
        iRowN = ftWidget._qcbSelectSlice.findText("row n")
        assert iRowN >= 0
        ftWidget._qcbSelectSlice.setCurrentIndex(iRowN)
        self.assertEqual(self._reconsParams["FT"]["FIXEDSLICE"], "4")
        self.assertEqual(self._reconsParams.to_dict()["FT"]["FIXEDSLICE"], "4")

        assert self._reconsParams["FT"]["VOLOUTFILE"] == 0
        self._reconsParams["FT"]["VOLOUTFILE"] = 0
        ftWidget._setSingleVolOrStack(1)
        self.assertTrue(self._reconsParams["FT"]["VOLOUTFILE"] == 1)

        assert self._reconsParams["FT"]["CORRECT_SPIKES_THRESHOLD"] == 0.04
        ftWidget._setSpikeThreshold(1.12)
        ftWidget._qleThresholdSpikesRm.editingFinished.emit()
        self.assertTrue(self._reconsParams["FT"]["CORRECT_SPIKES_THRESHOLD"] == 1.12)
        ftWidget._grpThreshold.setChecked(False)
        self.assertTrue(self._reconsParams["FT"]["CORRECT_SPIKES_THRESHOLD"] == "Inf")

        assert self._reconsParams["FT"]["RINGSCORRECTION"] == 0
        ftWidget._qcbRingsCorrection.setChecked(True)
        self.assertTrue(self._reconsParams["FT"]["RINGSCORRECTION"] == 1)

        assert self._reconsParams["FT"]["VOLSELECTION_REMEMBER"] == 0
        ftWidget._qcbVolSelRemenber.setChecked(True)
        self.assertTrue(self._reconsParams["FT"]["VOLSELECTION_REMEMBER"] == 1)

        assert self._reconsParams.to_dict()["FT"]["VOLSELECT"] == "total"
        assert self._reconsParams["FT"]["VOLSELECT"] == VolSelMode.total
        iSel = ftWidget._qcbVolumeSelection.findText("manual")
        assert iSel >= 0
        ftWidget._qcbVolumeSelection.setCurrentIndex(iSel)
        assert self._reconsParams.to_dict()["FT"]["VOLSELECT"] == "manual"
        assert self._reconsParams["FT"]["VOLSELECT"] == VolSelMode.manual

    def testFTAxisRead(self):
        ftAxisWidget = self.widget.getReconsParamSetEditor()._axisWidget

        assert self._reconsParams["FTAXIS"]["DO_AXIS_CORRECTION"] == 1
        self._reconsParams["FTAXIS"]["DO_AXIS_CORRECTION"] = 0
        self.qapp.processEvents()
        self.assertFalse(ftAxisWidget._doAxisCorrection.isChecked())

        assert self._reconsParams["FTAXIS"]["USE_TOMWER_AXIS"] == 1
        self._reconsParams["FTAXIS"]["USE_TOMWER_AXIS"] = 0
        self.assertFalse(ftAxisWidget._useTomwerAxis.isChecked())

        assert self._reconsParams["FTAXIS"]["TRY_USE_OLD_TOMWER_AXIS"] == 1
        self._reconsParams["FTAXIS"]["TRY_USE_OLD_TOMWER_AXIS"] = 0
        self.assertFalse(ftAxisWidget._useOldTomwerAxis.isChecked())

    def testDisplayRead(self):
        ftAxisWidget = self.widget.getReconsParamSetEditor()._displayWidget

        assert self._reconsParams["FT"]["SHOWPROJ"] == 0
        self._reconsParams["FT"]["SHOWPROJ"] = True
        self.qapp.processEvents()
        self.assertTrue(ftAxisWidget._qcbShowProj.isChecked())

        assert self._reconsParams["FT"]["SHOWSLICE"] == 1
        self._reconsParams["FT"]["SHOWSLICE"] = 0
        self.assertFalse(ftAxisWidget._qcbShowSlice.isChecked())

        assert self._reconsParams["FT"]["ANGLE_OFFSET"] == 0
        assert self._reconsParams["FT"]["ANGLE_OFFSET_VALUE"] == 0.0
        self._reconsParams["FT"]["ANGLE_OFFSET_VALUE"] = 2.3
        self.assertTrue(ftAxisWidget._qleAngleOffset.text() == "2.3")

    def testDisplayWrite(self):
        displayWidget = self.widget.getReconsParamSetEditor()._displayWidget
        assert self._reconsParams["FT"]["SHOWPROJ"] is False
        assert self._reconsParams.to_dict()["FT"]["SHOWPROJ"] == 0
        displayWidget._qcbShowProj.setChecked(True)
        self.assertTrue(self._reconsParams.to_dict()["FT"]["SHOWPROJ"] == 1)
        self.assertTrue(self._reconsParams["FT"]["SHOWPROJ"] is True)

        assert self._reconsParams.to_dict()["FT"]["SHOWSLICE"] == 1
        assert self._reconsParams["FT"]["SHOWSLICE"] is True
        displayWidget._qcbShowSlice.setChecked(False)
        self.assertTrue(self._reconsParams.to_dict()["FT"]["SHOWSLICE"] == 0)
        self.assertTrue(self._reconsParams["FT"]["SHOWSLICE"] is False)

        assert self._reconsParams["FT"]["ANGLE_OFFSET"] is False
        assert self._reconsParams["FT"]["ANGLE_OFFSET_VALUE"] == 0.0
        displayWidget._qleAngleOffset.setText("2.3")
        self.qapp.processEvents()
        displayWidget._qleAngleOffset.editingFinished.emit()
        assert self._reconsParams["FT"]["ANGLE_OFFSET_VALUE"] == 2.3
        assert self._reconsParams.to_dict()["FT"]["ANGLE_OFFSET"] == 1.0
        displayWidget._qleAngleOffset.setText("0.0")
        displayWidget._qleAngleOffset.editingFinished.emit()
        assert self._reconsParams.to_dict()["FT"]["ANGLE_OFFSET"] == 0
        assert self._reconsParams.to_dict()["FT"]["ANGLE_OFFSET_VALUE"] == 0.0

    def testPaganinRead(self):
        pagWidget = self.widget.getReconsParamSetEditor()._PaganinWidget

        assert self._reconsParams.to_dict()["PAGANIN"]["MODE"] == 0
        assert self._reconsParams["PAGANIN"]["MODE"] == PaganinMode.off
        self._reconsParams["PAGANIN"]["MODE"] = PaganinMode.multi
        self.assertTrue(pagWidget._qcbpaganin.currentText() == "multi")

        assert self._reconsParams["PAGANIN"]["DB"] == 500.0
        self._reconsParams["PAGANIN"]["DB"] = 2.3
        self.assertTrue(pagWidget._qleSigmaBeta.text() == "2.3")

        assert self._reconsParams["PAGANIN"]["DB2"] == 100.0
        self._reconsParams["PAGANIN"]["DB2"] = 3.6
        self.assertTrue(pagWidget._qleSigmaBeta2.text() == "3.6")

        assert self._reconsParams["PAGANIN"]["UNSHARP_COEFF"] == 3.0
        self._reconsParams["PAGANIN"]["UNSHARP_COEFF"] = 2.6
        self.assertTrue(pagWidget._unsharp_sigma_coeff.text() == "2.6")

        assert self._reconsParams["PAGANIN"]["UNSHARP_SIGMA"] == 0.8
        self._reconsParams["PAGANIN"]["UNSHARP_SIGMA"] = 1.0
        self.assertTrue(pagWidget._unsharp_sigma_mask_value.text() == "1.0")

        assert self._reconsParams["PAGANIN"]["THRESHOLD"] == 500.0
        self._reconsParams["PAGANIN"]["THRESHOLD"] = 1.0
        self.assertTrue(pagWidget._qleThreshold.text() == "1.0")

        assert self._reconsParams["PAGANIN"]["DILATE"] == 2
        self._reconsParams["PAGANIN"]["DILATE"] = 1
        self.assertTrue(pagWidget._qleDilatation.text() == "1")

        assert self._reconsParams["PAGANIN"]["MEDIANR"] == 4
        self._reconsParams["PAGANIN"]["MEDIANR"] = 1
        self.assertTrue(pagWidget._qleMedianFilterSize.text() == "1")

        assert self._reconsParams["PAGANIN"]["MKEEP_BONE"] == 0
        self._reconsParams["PAGANIN"]["MKEEP_BONE"] = 1
        self.assertTrue(pagWidget._qcbKeepBone.isChecked())

        assert self._reconsParams["PAGANIN"]["MKEEP_SOFT"] == 0
        self._reconsParams["PAGANIN"]["MKEEP_SOFT"] = 1
        self.assertTrue(pagWidget._qcbKeepSoft.isChecked())

        assert self._reconsParams["PAGANIN"]["MKEEP_ABS"] == 0
        self._reconsParams["PAGANIN"]["MKEEP_ABS"] = 1
        self.assertTrue(pagWidget._qcbKeepAbs.isChecked())

        assert self._reconsParams["PAGANIN"]["MKEEP_CORR"] == 0
        self._reconsParams["PAGANIN"]["MKEEP_CORR"] = 1
        self.assertTrue(pagWidget._qcbKeepCorr.isChecked())

        assert self._reconsParams["PAGANIN"]["MKEEP_MASK"] == 0
        self._reconsParams["PAGANIN"]["MKEEP_MASK"] = 1
        self.assertTrue(pagWidget._qcbKeepMask.isChecked())

    def testPaganinWrite(self):
        pagWidget = self.widget.getReconsParamSetEditor()._PaganinWidget

        iMulti = pagWidget._qcbpaganin.findText("multi")
        assert iMulti >= 0
        pagWidget._qcbpaganin.setCurrentIndex(iMulti)
        self.qapp.processEvents()
        self.assertTrue(self._reconsParams.to_dict()["PAGANIN"]["MODE"] == 3)
        self.assertTrue(self._reconsParams["PAGANIN"]["MODE"] == PaganinMode.multi)

        assert self._reconsParams["PAGANIN"]["DB"] == 500.0
        pagWidget._qleSigmaBeta.setText("2.3")
        pagWidget._qleSigmaBeta.editingFinished.emit()
        self.assertEqual(self._reconsParams.to_dict()["PAGANIN"]["DB"], "2.3")
        self.assertEqual(self._reconsParams["PAGANIN"]["DB"], 2.3)

        assert self._reconsParams["PAGANIN"]["DB2"] == 100.0
        pagWidget._qleSigmaBeta2.setText("3.6")
        pagWidget._qleSigmaBeta2.editingFinished.emit()
        self.assertEqual(self._reconsParams.to_dict()["PAGANIN"]["DB2"], "3.6")
        self.assertEqual(self._reconsParams["PAGANIN"]["DB2"], 3.6)

        assert self._reconsParams.to_dict()["PAGANIN"]["UNSHARP_COEFF"] == 3.0
        pagWidget._unsharp_sigma_coeff.setText("1.2")
        pagWidget._unsharp_sigma_coeff.editingFinished.emit()
        assert self._reconsParams.to_dict()["PAGANIN"]["UNSHARP_COEFF"] == 1.2

        assert self._reconsParams.to_dict()["PAGANIN"]["UNSHARP_SIGMA"] == 0.8
        pagWidget._unsharp_sigma_mask_value.setText("1.0")
        pagWidget._unsharp_sigma_mask_value.editingFinished.emit()
        assert self._reconsParams.to_dict()["PAGANIN"]["UNSHARP_SIGMA"] == 1.0

        assert self._reconsParams.to_dict()["PAGANIN"]["THRESHOLD"] == 500.0
        pagWidget._qleThreshold.setText("1.0")
        pagWidget._qleThreshold.editingFinished.emit()
        assert self._reconsParams.to_dict()["PAGANIN"]["THRESHOLD"] == 1.0

        assert self._reconsParams["PAGANIN"]["DILATE"] == 2
        pagWidget._qleDilatation.setText("1")
        pagWidget._qleDilatation.editingFinished.emit()
        assert self._reconsParams["PAGANIN"]["DILATE"] == 1

        assert self._reconsParams["PAGANIN"]["MEDIANR"] == 4
        pagWidget._qleMedianFilterSize.setText("1")
        pagWidget._qleMedianFilterSize.editingFinished.emit()
        assert self._reconsParams["PAGANIN"]["MEDIANR"] == 1

        assert self._reconsParams["PAGANIN"]["MKEEP_BONE"] == 0
        pagWidget._qcbKeepBone.setChecked(True)
        assert self._reconsParams["PAGANIN"]["MKEEP_BONE"] == 1

        assert self._reconsParams["PAGANIN"]["MKEEP_SOFT"] == 0
        pagWidget._qcbKeepSoft.setChecked(True)
        assert self._reconsParams["PAGANIN"]["MKEEP_SOFT"] == 1

        assert self._reconsParams["PAGANIN"]["MKEEP_ABS"] == 0
        pagWidget._qcbKeepAbs.setChecked(True)
        assert self._reconsParams["PAGANIN"]["MKEEP_ABS"] == 1

        assert self._reconsParams["PAGANIN"]["MKEEP_CORR"] == 0
        pagWidget._qcbKeepCorr.setChecked(True)
        assert self._reconsParams["PAGANIN"]["MKEEP_CORR"] == 1

        assert self._reconsParams["PAGANIN"]["MKEEP_MASK"] == 0
        pagWidget._qcbKeepMask.setChecked(True)
        assert self._reconsParams["PAGANIN"]["MKEEP_MASK"] == 1

    def testPyHSTRead(self):
        pyHSTWidget = self.widget.getReconsParamSetEditor()._PyHSTWidget
        self.assertEqual(
            type(self._reconsParams.pyhst), type(pyHSTWidget._recons_params)
        )
        self.assertTrue(id(self._reconsParams.pyhst) == id(pyHSTWidget._recons_params))

        assert self._reconsParams["PYHSTEXE"]["VERBOSE"] == 0
        self.assertFalse(pyHSTWidget._qcbverbose.isChecked())
        self.assertTrue(id(self._reconsParams.pyhst) == id(pyHSTWidget._recons_params))
        self._reconsParams["PYHSTEXE"]["VERBOSE"] = True
        self.assertTrue(id(self._reconsParams.pyhst) == id(pyHSTWidget._recons_params))

        self.qapp.processEvents()
        self.assertTrue(pyHSTWidget._qcbverbose.isChecked())

        assert self._reconsParams["PYHSTEXE"]["EXE"] == "pyhst2"
        self._reconsParams["PYHSTEXE"]["EXE"] = "pyhst3"
        self.assertTrue(pyHSTWidget._qcbPyHSTVersion.currentText() == "pyhst3")

        assert self._reconsParams["PYHSTEXE"]["OFFV"] == "pyhst2"
        self._reconsParams["PYHSTEXE"]["OFFV"] = "pyhst3"
        self.assertTrue(pyHSTWidget._qlOfficalVersion.text() == "pyhst3")

        assert self._reconsParams["PYHSTEXE"]["VERBOSE_FILE"] == "pyhst_out.txt"
        self._reconsParams["PYHSTEXE"]["VERBOSE_FILE"] = "outToto.dsad"
        self.assertTrue(pyHSTWidget._qleVerboseFile.text() == "outToto.dsad")

    def testPyHSTWrite(self):
        pyHSTWidget = self.widget.getReconsParamSetEditor()._PyHSTWidget
        assert self._reconsParams["PYHSTEXE"]["VERBOSE"] is False
        pyHSTWidget._qcbverbose.setChecked(True)
        self.qapp.processEvents()
        assert self._reconsParams["PYHSTEXE"]["VERBOSE"] is True

        assert self._reconsParams["PYHSTEXE"]["EXE"] == "pyhst2"
        pyHSTWidget._qcbPyHSTVersion.addItem("pyhst3Toto")
        pyHSTWidget._qcbPyHSTVersion.setCurrentIndex(
            pyHSTWidget._qcbPyHSTVersion.count() - 1
        )
        assert self._reconsParams["PYHSTEXE"]["EXE"] == "pyhst3Toto"

        assert self._reconsParams["PYHSTEXE"]["VERBOSE_FILE"] == "pyhst_out.txt"
        pyHSTWidget._qleVerboseFile.setText("outFile.dsad")
        pyHSTWidget._qleVerboseFile.editingFinished.emit()
        assert self._reconsParams["PYHSTEXE"]["VERBOSE_FILE"] == "outFile.dsad"

    def testBeamGeoRead(self):
        beamGeoWidget = self.widget.getReconsParamSetEditor()._beamGeoWidget

        assert self._reconsParams.to_dict()["BEAMGEO"]["TYPE"] == "p"
        assert self._reconsParams["BEAMGEO"]["TYPE"] == BeamGeoType.parallel
        self._reconsParams["BEAMGEO"]["TYPE"] = "f"
        self.assertTrue(beamGeoWidget._qcbType.currentText() == "fan beam")

        assert self._reconsParams["BEAMGEO"]["SX"] == 0.0
        self._reconsParams["BEAMGEO"]["SX"] = 10.2
        assert self._reconsParams["BEAMGEO"]["SX"] == 10.2
        self.assertTrue(beamGeoWidget._qleSX.text() == "10.2")

        assert self._reconsParams["BEAMGEO"]["SY"] == 0.0
        self._reconsParams["BEAMGEO"]["SY"] = 20.2
        self.assertTrue(beamGeoWidget._qleSY.text() == "20.2")

        assert self._reconsParams["BEAMGEO"]["DIST"] == 55.0
        self._reconsParams["BEAMGEO"]["DIST"] = 1.0
        self.assertTrue(beamGeoWidget._qleDIST.text() == "1.0")

    def testBeamGeoWrite(self):
        beamGeoWidget = self.widget.getReconsParamSetEditor()._beamGeoWidget

        assert self._reconsParams["BEAMGEO"]["TYPE"] == BeamGeoType.parallel
        assert self._reconsParams.to_dict()["BEAMGEO"]["TYPE"] == "p"
        iType = beamGeoWidget._qcbType.findText("fan beam")
        assert iType >= 0
        beamGeoWidget._qcbType.setCurrentIndex(iType)
        assert self._reconsParams["BEAMGEO"]["TYPE"] == BeamGeoType.fan
        assert self._reconsParams.to_dict()["BEAMGEO"]["TYPE"] == "f"

        assert self._reconsParams["BEAMGEO"]["SX"] == 0.0
        beamGeoWidget._qleSX.setText("10.2")
        beamGeoWidget._qleSX.editingFinished.emit()
        assert self._reconsParams["BEAMGEO"]["SX"] == 10.2

        assert self._reconsParams["BEAMGEO"]["SY"] == 0.0
        beamGeoWidget._qleSY.setText("11.2")
        beamGeoWidget._qleSY.editingFinished.emit()
        assert self._reconsParams["BEAMGEO"]["SY"] == 11.2

        assert self._reconsParams["BEAMGEO"]["DIST"] == 55.0
        beamGeoWidget._qleDIST.setText("0.2")
        beamGeoWidget._qleDIST.editingFinished.emit()
        assert self._reconsParams["BEAMGEO"]["DIST"] == 0.2

    def testExpertRead(self):
        expertWidget = self.widget.getReconsParamSetEditor()._expertWidget

        assert self._reconsParams["FT"]["NUM_PART"] == 4
        self._reconsParams["FT"]["NUM_PART"] = 2
        self.assertTrue(expertWidget._qsbNumPart.value() == 2)

        assert self._reconsParams["FT"]["VERSION"] == "fastomo3 3.2"
        self._reconsParams["FT"]["VERSION"] = "toto"
        self.assertTrue(expertWidget._qleVersion.text() == "toto")

        assert self._reconsParams["FT"]["DATABASE"] == 0
        self._reconsParams["FT"]["DATABASE"] = 1
        self.assertTrue(expertWidget._qcbDataBase.isChecked())

        assert self._reconsParams["FT"]["NO_CHECK"] == 0
        self.assertFalse(expertWidget._qcbNocheck.isChecked())
        self._reconsParams["FT"]["NO_CHECK"] = 1
        self.assertTrue(expertWidget._qcbNocheck.isChecked())

        assert self._reconsParams["FT"]["ZEROOFFMASK"] == 1
        self._reconsParams["FT"]["ZEROOFFMASK"] = 0
        self.assertFalse(expertWidget._qcbZeroRegionMask.isChecked())

        assert self._reconsParams["FT"]["FIXHD"] == 0
        self._reconsParams["FT"]["FIXHD"] = 1
        self.assertTrue(expertWidget._qcbFixHD.isChecked())

    def testExpertWrite(self):
        expertWidget = self.widget.getReconsParamSetEditor()._expertWidget

        assert self._reconsParams["FT"]["NUM_PART"] == 4
        expertWidget._qsbNumPart.setValue(2)
        self.qapp.processEvents()
        assert self._reconsParams["FT"]["NUM_PART"] == 2

        assert self._reconsParams["FT"]["DATABASE"] == 0
        expertWidget._qcbDataBase.setChecked(True)
        assert self._reconsParams["FT"]["DATABASE"] == 1

        assert self._reconsParams["FT"]["NO_CHECK"] == 0
        expertWidget._qcbNocheck.setChecked(True)
        assert self._reconsParams["FT"]["NO_CHECK"] == 1

        assert self._reconsParams["FT"]["ZEROOFFMASK"] == 1
        expertWidget._qcbZeroRegionMask.setChecked(False)
        assert self._reconsParams["FT"]["ZEROOFFMASK"] == 0

        assert self._reconsParams["FT"]["FIXHD"] == 0
        expertWidget._qcbFixHD.setChecked(True)
        assert self._reconsParams["FT"]["FIXHD"] == 1

    def testOtherRead(self):
        assert isinstance(self._reconsParams["FT"], _ReconsParam)
        self._reconsParams["FT"]["MYPARAM"] = "tata"
        assert isinstance(self._reconsParams["FT"], _ReconsParam)
        otherWidget = self.widget.getReconsParamSetEditor()._otherWidget
        widgetTata = otherWidget.paramToWidget["FT"]["MYPARAM"]
        assert widgetTata.text() == "tata"
        p = self._reconsParams["FT"]
        p._set_parameter_value("MYPARAM", "toto")
        self._reconsParams["FT"]._set_parameter_value("MYPARAM", "toto")
        self.qapp.processEvents()
        otherWidget = self.widget.getReconsParamSetEditor()._otherWidget
        widgetTata = otherWidget.paramToWidget["FT"]["MYPARAM"]
        self.assertTrue(widgetTata.text() == "toto")

    def testOtherWrite(self):
        self._reconsParams["FT"]["MYPARAM"] = "tata"
        otherWidget = self.widget.getReconsParamSetEditor()._otherWidget
        widgetTata = otherWidget.paramToWidget["FT"]["MYPARAM"]
        widgetTata.setText("toto")
        self._reconsParams["FT"]["MYPARAM"] = "toto"


def suite():
    test_suite = unittest.TestSuite()
    for ui in (
        TestControlsFTSeriesWidget,
        TestFTSerieWidget,
        TestExploreDatasetForH5,
        TestLoadSaveH5File,
        TestSync,
    ):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "16/06/2021"


import numpy
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt
import pytest
from tomwer.gui.reconstruction.ftserie.reconsparamseditor import ReconsParamSetEditor
from tomwer.test.utils import skip_gui_test
import gc


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestReconsParamSet(TestCaseQt):
    """
    Test the ReconsParamSetEditor.
    Make sure we can iterate over the set of ReconsParams
    """

    def setUp(self):
        super().setUp()
        self.widget = ReconsParamSetEditor(parent=None)

    def tearDown(self):
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        self.qapp.processEvents()
        gc.collect()
        super().tearDown()

    def testSeveralPaganin(self):
        iMulti = self.widget._PaganinWidget._qcbpaganin.findText("multi")
        assert iMulti >= 0
        self.widget._PaganinWidget._qcbpaganin.setCurrentIndex(iMulti)
        self.widget._PaganinWidget._qleSigmaBeta.setText("1, 2, 3")
        self.widget._PaganinWidget._qleSigmaBeta.editingFinished.emit()
        self.widget._PaganinWidget._qleSigmaBeta2.setText("0:10:2")
        self.widget._PaganinWidget._qleSigmaBeta2.editingFinished.emit()
        qt.qApp.processEvents()
        reconsParamList = self.widget.getReconsParamSet()
        assert len(reconsParamList) == (5 * 3)
        combinations = []
        for db in (1, 2, 3):
            num = 5
            for db2 in numpy.linspace(0, 10, num=num, endpoint=True):
                combinations.append((db, db2))
        for param in reconsParamList:
            c = (param["PAGANIN"]["DB"], param["PAGANIN"]["DB2"])
            assert c in combinations
            combinations.remove(c)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "25/02/2019"


import numpy
import logging
import enum
from numpy.linalg import inv

_logger = logging.getLogger(__file__)
try:
    from scipy.ndimage import shift as shift_scipy

    has_scipy_shift = True
except ImportError:
    has_scipy_shift = False
    _logger.info("no scipy.ndimage.shift detected, will use numpy.fft instead")


def shift_img(
    data: numpy.ndarray, dx: float, dy: float, cval: float = 0.0
) -> numpy.ndarray:
    """
    Apply simple 2d image shift in 'constant mode'.

    :param data:
    :type data: numpy.ndarray
    :param dx: x translation to be applied
    :type dx: float
    :param dy: y translation to be applied
    :type dy: float
    :param float cval: value to replace the shifted values

    :return: shifted image
    :rtype: numpy.ndarray
    """
    assert data.ndim == 2
    assert dx is not None
    assert dy is not None
    _logger.debug("apply shift dx=%s, dy=%s " % (dx, dy))

    if has_scipy_shift:
        return shift_scipy(
            input=data, shift=(dy, dx), order=1, mode="constant", cval=cval
        )
    else:
        ynum, xnum = data.shape
        xmin = int(-numpy.fix(xnum / 2))
        xmax = int(numpy.ceil(xnum / 2) - 1)
        ymin = int(-numpy.fix(ynum / 2))
        ymax = int(numpy.ceil(ynum / 2) - 1)

        nx, ny = numpy.meshgrid(
            numpy.linspace(xmin, xmax, xnum), numpy.linspace(ymin, ymax, ynum)
        )
        # cast variables to float
        ny = numpy.asarray(ny, numpy.float32)
        nx = numpy.asarray(nx, numpy.float32)
        res = abs(
            numpy.fft.ifft2(
                numpy.fft.fft2(data)
                * numpy.exp(1.0j * 2.0 * numpy.pi * (-dy * ny / ynum + -dx * nx / xnum))
            )
        )

        # apply constant filter
        if dx > 0:
            res[:, 0 : int(numpy.ceil(dx))] = cval
        elif dx < 0:
            res[:, xnum + int(numpy.ceil(dx)) :] = cval
        return res


class ImageScaleMethod(enum.Enum):
    RAW = "raw"
    MEAN = "mean"
    MEDIAN = "median"


def scale_img2_to_img1(img_1, img_2, method=ImageScaleMethod.MEAN):
    """
    scale image2 relative to image 1 in such a way they have same min and
    max. Scale will be apply from and to 'data' / raw data

    :param img_1: reference image
    :type: numpy.array
    :param numpy.array img_2: image to scale
    :type: numpy.array
    :param method: method to apply scaling
    :type: ImageScaleMethod
    :return:
    """
    assert method in ImageScaleMethod
    assert img_1.ndim == 2
    assert img_2.shape == img_1.shape
    min1 = img_2.min()
    max1 = img_2.max()
    min0 = img_1.min()
    max0 = img_1.max()

    if method is ImageScaleMethod.RAW:
        a = (min0 - max0) / (min1 - max1)
        b = (min1 * max0 - min0 * max1) / (min1 - max1)
        return a * img_2 + b
    else:
        if method is ImageScaleMethod.MEAN:
            me0 = img_1.mean()
            me1 = img_2.mean()
        elif method is ImageScaleMethod.MEDIAN:
            me0 = img_1.median()
            me1 = img_2.median()
        else:
            raise ValueError("method not managed", method)

        vec0 = numpy.mat([[min0], [me0], [max0]])
        matr = numpy.mat(
            [[min1 * min1, min1, 1.0], [me1 * me1, me1, 1.0], [max1 * max1, max1, 1.0]]
        )
        vec1 = inv(matr) * vec0
        return (
            float(vec1[0]) * (img_2 * img_2) + float(vec1[1]) * img_2 + float(vec1[2])
        )



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "23/07/2020"


from silx.gui import qt
from tomwer.gui.visualization.dataviewer import DataViewer
from tomwer.core.process.control.scanvalidator import ScanValidator as BaseDataValidator
from tomwer.gui.utils.waiterthread import QWaiterThread
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanbase import _TomwerBaseDock
from tomwer.gui.qfolderdialog import QScanDialog
from tomwer.core.scan.scanfactory import ScanFactory
import weakref
from contextlib import AbstractContextManager
import logging
import time

_logger = logging.getLogger(__file__)


WAIT_TIME_MEM_REL = 20
"""Time (in sec) to wait to check if the scan validator has to release his
stack of scans. This is a security to make sure we are not keeping
unnecessary data during an acquisition in lbsram"""


class DataValidator(qt.QMainWindow, BaseDataValidator):
    """
    Widget used to validate an acquisition or recall some processing
    """

    sigScanReady = qt.Signal(TomwerScanBase)
    """signal emitted when a scan is validated"""

    sigChangeReconsParams = qt.Signal(_TomwerBaseDock)
    """signal emitted when the user request modifications on reconstruction
    parameters"""

    class SliderCM(AbstractContextManager):
        """Simple context manager to hida / show button dialogs"""

        def __init__(self, slider):
            self._slider = slider
            self._old = None

        def __enter__(self):
            self._old = self._slider.blockSignals(True)

        def __exit__(self, exc_type, exc_val, exc_tb):
            self._slider.blockSignals(self._old)

    def __init__(self, parent, timeout_loop_memory=WAIT_TIME_MEM_REL):
        qt.QMainWindow.__init__(
            self, parent, memoryReleaser=QWaiterThread(timeout_loop_memory)
        )
        # needed on lbs191... otherwise crash with PyQt5.11. with a QThread destroyed...
        time.sleep(0.01)
        BaseDataValidator.__init__(
            self, memoryReleaser=QWaiterThread(timeout_loop_memory)
        )
        self._centralWidget = DataViewer(parent=parent)
        self.setCentralWidget(self._centralWidget)

        # scan slider and add scan
        self._lateralWidget = qt.QWidget(self)
        self._lateralWidget.setLayout(qt.QVBoxLayout())
        self._addScanButton = qt.QPushButton(self)
        style = qt.QApplication.style()
        icon = style.standardIcon(qt.QStyle.SP_DirIcon)
        self._addScanButton.setIcon(icon)
        self._lateralWidget.layout().addWidget(self._addScanButton)
        self._scanSelectorSlider = _ScanSelectorWidget(parent=self)
        self._lateralWidget.layout().addWidget(self._scanSelectorSlider)
        self._lateralWidget.layout().setContentsMargins(0, 0, 0, 0)

        self._scanSelectorSliderDocker = qt.QDockWidget(self)
        self._scanSelectorSliderDocker.setWidget(self._lateralWidget)
        self._scanSelectorSliderDocker.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self._scanSelectorSliderDocker)

        # control buttons
        self._controlButtons = _DataControlWidget(parent=self)
        self._controlButtonsDocker = qt.QDockWidget(self)
        self._controlButtonsDocker.setWidget(self._controlButtons)
        self._controlButtonsDocker.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self.addDockWidget(qt.Qt.BottomDockWidgetArea, self._controlButtonsDocker)

        self._scanSelectorSlider.setMinimum(0)
        self._scanSelectorSlider.setMaximum(0)

        # connect signal / slot
        self._scanSelectorSlider.sigScanChanged.connect(self._updateCurrentScan)
        _controlButtons = self._controlButtons
        _controlButtons.sigValidateScan.connect(self.validateCurrentScan)
        _controlButtons.sigChangeReconstructionParametersScan.connect(
            self.changeReconsParamCurrentScan
        )
        _controlButtons.toggled.connect(self.setManualValidation)
        self._addScanButton.pressed.connect(self._addNewScanCallback)

    def close(self) -> bool:
        self._centralWidget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._centralWidget.close()
        self._centralWidget = None
        qt.QMainWindow.close(self)

    def addScan(self, scan):
        """

        :param scan:
        :return:
        """
        assert isinstance(scan, TomwerScanBase)
        BaseDataValidator.addScan(self, scan)

        with self.SliderCM(self._scanSelectorSlider):
            current_index = self._scanSelectorSlider.value()
            self._scanSelectorSlider.setRange(0, len(self._scansToValidate) - 1)
            self._scanSelectorSlider.setValue(current_index)

        if str(scan) in self._scansToValidate:
            self.setActiveScan(scan)

    def setActiveScan(self, scan: TomwerScanBase):
        """

        :param scan:
        :return:
        """
        if scan is None:
            self.clear()
        elif str(scan) not in self._scansToValidate:
            _logger.warning("scan has not been registered")
        else:
            with self.SliderCM(self._scanSelectorSlider):
                self._centralWidget.setScan(scan)
                self._controlButtons.setScan(scan)
                index = list(self._scansToValidate.keys()).index(str(scan))
                self._scanSelectorSlider.setValue(index)

    def _updateCurrentScan(self):
        with self.SliderCM(self._scanSelectorSlider):
            self.setActiveScan(self.getCurrentScan())

    def getCurrentScan(self) -> TomwerScanBase:
        """

        :return:
        """
        current_scan_index = self._scanSelectorSlider.value()
        if current_scan_index >= len(self._scansToValidate):
            return None
        else:
            scan_id = list(self._scansToValidate.keys())[current_scan_index]
            scan = self._scansToValidate[scan_id]
            assert isinstance(scan, TomwerScanBase)
            return scan

    def validateCurrentScan(self):
        scan = self.getCurrentScan()
        if scan is None:
            return
        else:
            BaseDataValidator._validateScan(self, scan)
            with self.SliderCM(self._scanSelectorSlider):
                current_index = self._scanSelectorSlider.value()
                self._scanSelectorSlider.setRange(0, len(self._scansToValidate) - 1)
                # note: slider force the value to be in [minimum, maximum]
                self._scanSelectorSlider.setValue(current_index)
                self._updateCurrentScan()

    def changeReconsParamCurrentScan(self):
        current_scan = self.getCurrentScan()
        if current_scan:
            assert isinstance(current_scan, TomwerScanBase)
            BaseDataValidator._changeReconsParam(self, current_scan)

    def _sendScanReady(self, scan):
        if scan is not None:
            assert isinstance(scan, TomwerScanBase)
            self.sigScanReady.emit(scan)

    def _sendUpdateReconsParam(self, scan):
        if scan is not None:
            assert isinstance(scan, _TomwerBaseDock)
            self.sigChangeReconsParams.emit(scan)

    def clear(self):
        try:
            self._centralWidget.clear()
        except AttributeError:
            # if has been removed before calling clear
            pass

    def _validateStack(self, filter_=None):
        BaseDataValidator._validateStack(self, filter_=filter_)
        self._updateCurrentScan()

    def setAutomaticValidation(self, auto):
        self._controlButtons.setChecked(not auto)

    def _addNewScanCallback(self):
        """Callback when we want to add a new scan to the scan validator"""
        dialog = QScanDialog(self, multiSelection=True)

        if not dialog.exec_():
            dialog.close()
            return

        foldersSelected = dialog.files_selected()
        new_active_scan = None
        for folder in foldersSelected:
            try:
                scans = ScanFactory.create_scan_objects(scan_path=folder)
            except Exception as e:
                _logger.error(
                    "cannot create scan instances from {}. Error is {}".format(
                        folder, e
                    )
                )
            else:
                for scan in scans:
                    self.addScan(scan=scan)
                    new_active_scan = scan

        if new_active_scan is not None:
            self.setActiveScan(new_active_scan)


class _ScanSelectorWidget(qt.QWidget):
    """Slider to select the scan we want to 'check'"""

    sigScanChanged = qt.Signal(int)
    """signal emitted when the scan change. Value is the scan index"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QHBoxLayout())

        self._qslider = qt.QSlider(qt.Qt.Vertical, self)
        self.layout().addWidget(self._qslider)

        self._verticalLabel = _VerticalLabel(
            "stack of received scan", parent=self, revert=True
        )
        self.layout().addWidget(self._verticalLabel)

        # connect signal / slot
        self._qslider.valueChanged.connect(self._valueHasChanged)

    def _valueHasChanged(self, *args, **kwargs):
        self.sigScanChanged.emit(self._qslider.value())

    def value(self):
        return self._qslider.value()

    def setValue(self, value):
        self._qslider.setValue(value)

    def setMinimum(self, p_int):
        self._qslider.setMinimum(p_int)

    def setMaximum(self, p_int):
        self._qslider.setMaximum(p_int)

    def setRange(self, p_int, p_int_1):
        self._qslider.setRange(p_int, p_int_1)


class _DataControlWidget(qt.QGroupBox):
    """Set of buttons to control the curremt scan"""

    """
        Class containing all the validation buttons
        and sending signals when they are pushed

        :param QObject parent: the parent of the QTabWidget
        :param :class:`.FtserieReconstruction`: the scan to display
        """

    sigValidateScan = qt.Signal(str)
    sigCancelScan = qt.Signal(str)
    sigRedoAcquisitionScan = qt.Signal(TomwerScanBase)
    sigChangeReconstructionParametersScan = qt.Signal(TomwerScanBase)

    def __init__(self, parent=None):
        qt.QGroupBox.__init__(self, title="Validate manually", parent=parent)
        self._scan = None
        self.setCheckable(True)
        layout = qt.QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        #  validate button
        self.validateButton = qt.QPushButton("Validate")
        style = qt.QApplication.style()
        self.validateButton.setIcon(style.standardIcon(qt.QStyle.SP_DialogApplyButton))
        self.validateButton.pressed.connect(self.__validated)
        layout.addWidget(self.validateButton, 0, 2)

        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        layout.addWidget(spacer, 0, 1)

        #  change reconstruction parameters button
        self.changeReconsParamButton = qt.QPushButton(
            "Change reconstruction parameters"
        )
        self.changeReconsParamButton.setIcon(
            style.standardIcon(qt.QStyle.SP_FileDialogContentsView)
        )
        self.changeReconsParamButton.pressed.connect(
            self.__updateReconstructionParameters
        )
        layout.addWidget(self.changeReconsParamButton, 2, 0)

    def setScan(self, scan):
        self._scan = weakref.ref(scan)

    def setEnabled(self, b):
        self.validateButton.setEnabled(b)
        self.changeReconsParamButton.setEnabled(b)

    def __validated(self):
        """Callback when the validate button is pushed"""
        self.sigValidateScan.emit("")

    def __redoacquisition(self):
        """Callback when the redo acquisition button is pushed"""
        if self._scan is not None and self._scan() is not None:
            self.sigRedoAcquisitionScan.emit(self._scan())

    def __updateReconstructionParameters(self):
        """Callback when the change reconstruction button is pushed"""
        if self._scan is not None and self._scan() is not None:
            self.sigChangeReconstructionParametersScan.emit(self._scan())


class _VerticalLabel(qt.QLabel):
    """Display vertically the given text"""

    def __init__(self, text, parent=None, revert=False):
        """

        :param text: the legend
        :param parent: the Qt parent if any
        """
        qt.QLabel.__init__(self, text, parent)
        self.revert = revert
        self.setLayout(qt.QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

    def paintEvent(self, event):
        painter = qt.QPainter(self)
        painter.setFont(self.font())

        painter.translate(0, self.rect().height())
        painter.rotate(90)
        if self.revert:
            newRect = qt.QRect(
                -self.rect().height(),
                -self.rect().width(),
                self.rect().height(),
                self.rect().width(),
            )
        else:
            newRect = self.rect()

        painter.drawText(newRect, qt.Qt.AlignHCenter, self.text())

        fm = qt.QFontMetrics(self.font())
        preferedHeight = fm.width(self.text())
        preferedWidth = fm.height()
        self.setFixedWidth(preferedWidth)
        self.setMinimumHeight(preferedHeight)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "15/12/2021"


from silx.gui import qt
from tomwer.core.process.reconstruction.nabu.castvolume import (
    RESCALE_MAX_PERCENTILE,
    RESCALE_MIN_PERCENTILE,
    DataType,
    DEFAULT_OUTPUT_DIR,
)
from tomwer.core.process.reconstruction.nabu.nabucommon import NabuOutputFileFormat
from tomwer.gui.reconstruction.nabu.nabuconfig.output import QNabuFileFormatComboBox
import logging

_logger = logging.getLogger(__name__)


class CastVolumeWidget(qt.QWidget):

    sigConfigChanged = qt.Signal()
    """Signal emit when the configuration changed"""

    def __init__(self, parent) -> None:
        super().__init__(parent=parent)

        self.setLayout(qt.QGridLayout())

        # output data size
        self._castToLabel = qt.QLabel("cast to", self)
        self._castToLabel.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum)
        self.layout().addWidget(self._castToLabel, 0, 0, 1, 1)
        self._outputDataSizeCB = qt.QComboBox(self)
        for data_size in DataType:
            self._outputDataSizeCB.addItem(data_size.value)
        self.layout().addWidget(self._outputDataSizeCB, 0, 2, 1, 2)
        # output file format
        self._outputFileformatLabel = qt.QLabel("output file format", self)
        self.layout().addWidget(self._outputFileformatLabel, 1, 0, 1, 1)
        self._outputFileformatCB = QNabuFileFormatComboBox(self)
        for file_format in (
            NabuOutputFileFormat.EDF,
            NabuOutputFileFormat.HDF5,
            NabuOutputFileFormat.JP2K,
        ):
            idx = self._outputFileformatCB.findText(file_format.value)
            if idx >= 0:
                self._outputFileformatCB.view().setRowHidden(idx, True)
        self.layout().addWidget(self._outputFileformatCB, 1, 2, 1, 1)
        # rescale percentiles
        self._percentilesLabel = qt.QLabel("rescale percentiles")
        self.layout().addWidget(self._percentilesLabel, 2, 0, 1, 1)
        self._lowPercentileQSB = qt.QSpinBox(self)
        self._lowPercentileQSB.setRange(0, 100)
        self._lowPercentileQSB.setPrefix("min:")
        self._lowPercentileQSB.setValue(RESCALE_MIN_PERCENTILE)
        self.layout().addWidget(self._lowPercentileQSB, 2, 1, 1, 1)
        self._highPercentileQSB = qt.QSpinBox(self)
        self._highPercentileQSB.setRange(0, 100)
        self._highPercentileQSB.setPrefix("max:")
        self._highPercentileQSB.setValue(RESCALE_MAX_PERCENTILE)
        self.layout().addWidget(self._highPercentileQSB, 2, 2, 1, 1)
        # save dir
        self._saveDirLabel = qt.QLabel("output directory", self)
        self.layout().addWidget(self._saveDirLabel, 3, 0, 1, 1)
        self._useDefaultSaveDirQCB = qt.QCheckBox("default", self)
        self._useDefaultSaveDirQCB.setToolTip(
            f"Default directory is: {DEFAULT_OUTPUT_DIR}"
        )
        self._useDefaultSaveDirQCB.setSizePolicy(
            qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum
        )
        self._useDefaultSaveDirQCB.setChecked(True)
        self.layout().addWidget(self._useDefaultSaveDirQCB, 3, 1, 1, 1)
        self._saveDirQLE = qt.QLineEdit(DEFAULT_OUTPUT_DIR, self)
        self._saveDirQLE.setToolTip(
            """
            can contains pattern / keywords. Those should be provided as '{keyword}'. Here are pattern currently handled:
            \n - 'scan_dir_name': returns name of the directory containing the acquisition (! not a path !)"
            \n - 'scan_basename': returns basename of the directory containing the acquisition"
            \n - 'scan_parent_dir_basename': returns basename of the PARENT directory containing the acquisition"
            """
        )
        self._saveDirQLE.setVisible(False)
        self.layout().addWidget(self._saveDirQLE, 3, 2, 1, 1)
        # overwrite
        self._overwriteCB = qt.QCheckBox("overwrite", self)
        self._overwriteCB.setChecked(True)
        self.layout().addWidget(self._overwriteCB, 4, 0, 1, 1)
        # spacer
        self._spacer = qt.QWidget(self)
        self._spacer.setSizePolicy(
            qt.QSizePolicy.MinimumExpanding, qt.QSizePolicy.Expanding
        )
        self.layout().addWidget(self._spacer, 99, 0, 1, 3)

        # connect signal / slot
        self._outputFileformatCB.currentIndexChanged.connect(self._configChanged)
        self._outputDataSizeCB.currentIndexChanged.connect(self._configChanged)
        self._saveDirQLE.editingFinished.connect(self._configChanged)
        self._useDefaultSaveDirQCB.toggled.connect(self._configChanged)
        self._overwriteCB.toggled.connect(self._configChanged)
        self._useDefaultSaveDirQCB.toggled.connect(self._updateOutputDirVis)

    def getOutputDataSize(self) -> DataType:
        return DataType.from_value(self._outputDataSizeCB.currentText())

    def setOutputDataSize(self, data_size) -> None:
        data_size = DataType.from_value(data_size)
        idx = self._outputDataSizeCB.findText(data_size.value)
        if idx >= 0:
            self._outputDataSizeCB.setCurrentIndex(idx)

    def getOutputDir(self) -> str:
        if self._useDefaultSaveDirQCB.isChecked():
            return DEFAULT_OUTPUT_DIR
        else:
            return self._saveDirQLE.text()

    def setOutputDir(self, output_dir: str) -> None:
        self._useDefaultSaveDirQCB.setChecked(output_dir == DEFAULT_OUTPUT_DIR)
        self._saveDirQLE.setText(output_dir)

    def getOutputFileFormat(self) -> NabuOutputFileFormat:
        return NabuOutputFileFormat.from_value(self._outputFileformatCB.currentText())

    def setOutputFileformat(self, file_format: str) -> None:
        file_format = NabuOutputFileFormat.from_value(file_format)
        idx = self._outputFileformatCB.findText(file_format.value)
        if idx >= 0:
            self._outputFileformatCB.setCurrentIndex(idx)

    def getOverwrite(self) -> bool:
        return self._overwriteCB.isChecked()

    def setOverwrite(self, remove) -> None:
        self._overwriteCB.setChecked(remove)

    def getConfiguration(self) -> dict:
        return {
            "output_data_type": self.getOutputDataSize().value,
            "output_file_format": self.getOutputFileFormat().value,
            "output_dir": self.getOutputDir(),
            "overwrite": self.getOverwrite(),
            "rescale_percentiles": self.getRescalePercentiles(),
        }

    def getRescalePercentiles(self) -> tuple:
        return self._lowPercentileQSB.value(), self._highPercentileQSB.value()

    def setRescalePercentiles(self, low, high) -> tuple:
        self._lowPercentileQSB.setValue(low)
        self._highPercentileQSB.setValue(high)

    def setConfiguration(self, config: dict) -> None:
        output_data_type = config.get("output_data_type", None)
        if output_data_type is not None:
            self.setOutputDataSize(output_data_type)
        output_file_format = config.get("output_file_format", None)
        if output_file_format is not None:
            self.setOutputFileformat(output_file_format)
        output_dir = config.get("output_dir", None)
        if output_dir is not None:
            self.setOutputDir(output_dir)
        overwrite = config.get("overwrite", None)
        if overwrite is not None:
            self.setOverwrite(overwrite)
        rescale_percentiles = config.get("rescale_percentiles", None)
        if rescale_percentiles is not None:
            if not len(rescale_percentiles) == 2:
                _logger.error(
                    f"percentiles is expected to contain two values not {len(rescale_percentiles)}"
                )
            else:
                self.setRescalePercentiles(
                    rescale_percentiles[0], rescale_percentiles[1]
                )

    def _configChanged(self, *args, **kwargs):
        self.sigConfigChanged.emit()

    def _updateOutputDirVis(self, *args, **kwargs):
        self._saveDirQLE.setVisible(not self._useDefaultSaveDirQCB.isChecked())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from math import pi
from bisect import bisect
from ..utils import generate_powers
from silx.utils.enum import Enum as _Enum


def lmicron_to_db(Lmicron, energy, distance):
    """
    Utility to convert the "Lmicron" parameter of PyHST
    to a value of delta/beta.

    Parameters
    ----------
    Lmicron: float
        Length in microns, values of the parameter "PAGANIN_Lmicron"
        in PyHST2 parameter file.
    energy: float
        Energy in keV.
    distance: float
        Sample-detector distance in microns

    Formula
    -------
    The conversion is done using the formula

    $$
    L^2 = \pi \lambda D \frac{\delta}{\beta}  # noqa W605
    $$
    The PyHST2 normalization differs from the one used by other softwares
    like tomopy by a factor $1/(4\pi^2)$
    """
    L2 = Lmicron ** 2
    wavelength = 1.23984199e-3 / energy
    return L2 / (pi * wavelength * distance)


class PaddingMode(_Enum):
    ZEROS = "zeros"
    MEAN = "mean"
    EDGE = "edge"
    SYMMETRIC = "symmetric"
    REFLECT = "reflect"


class PaganinPhaseRetrieval(object):
    """
    Paganin Phase Retrieval for an infinitely distant point source.
    Formula (10) in [1].

    Parameters
    ----------
    shape: int or tuple
        Shape of each radio, in the format (num_rows, num_columns), i.e
        (size_vertical, size_horizontal).
        If an integer is provided, the shape is assumed to be square.
    distance : float, optional
        Propagation distance in cm.
    energy : float, optional
        Energy in keV.
    delta_beta: float, optional
        delta/beta ratio, where n = (1 - delta) + i*beta is the complex
        refractive index of the sample.
    pixel_size : float, optional
        Detector pixel size in microns.
    padding : str, optional
        Padding method. Available are "zeros", "mean", "edge", "sym",
        "reflect". Default is "edge".
        Please refer to the "Padding" section below for more details.
    margin: tuple, optional
        The user may provide integers values U, D, L, R as a tuple under the
        form ((U, D), (L, R)) (same syntax as numpy.pad()).
        The resulting filtered radio will have a size equal to
        (size_vertic - U - D, size_horiz - L - R).
        These values serve to create a "margin" for the filtering process,
        where U, D, L R are the margin of the Up, Down, Left and Right part,
        respectively.
        The filtering is done on a subset of the input radio. The subset
        size is (Nrows - U - D, Ncols - R - L).
        The margins is used to do the padding for the rest of the padded
        array.

        For example in one dimension, where padding="edge":

        <------------------------------ padded_size --------------------------->
        [padding=edge | padding=data | radio data | padding=data | padding=edge]
        <------ N2 ---><----- L -----><- (N-L-R)--><----- R -----><----- N2 --->

        Some or all the values U, D, L, R can be 0. In this case,
        the padding of the parts related to the zero values will
        fall back to the one of "padding" parameter.
        For example, if padding="edge" and L, R are 0, then
        the left and right parts will be padded with the edges, while
        the Up and Down parts will be padded using the the user-provided
        margins of the radio, and the final data will have shape
        (Nrows - U - D, Ncols).
        Some or all the values U, D, L, R can be the string "auto".
        In this case, the values of U, D, L, R are automatically computed
        as a function of the Paganin filter width.
    use_R2C: bool, optional
        Whether to use Real-to-Complex (R2C) transform instead of
        standard Complex-to-Complex transform, providing better performances

    Padding methods
    ---------------
    The phase retrieval is a convolution done in Fourier domain using FFT,
    so the Fourier transform size has to be at least twice the size of
    the original data. Mathematically, the data should be padded with zeros
    before being Fourier transformed. However, in practice, this can lead
    to artefacts at the edges (Gibbs effect) if the data does not go to
    zero at the edges.
    Apart from applying an apodization (Hamming, Blackman, etc), a common
    strategy to avoid these artefacts is to pad the data.
    In tomography reconstruction, this is usually done by replicating the
    last(s) value(s) of the edges ; but one can think of other methods:

       - "zeros": the data is simply padded with zeros.
       - "mean": the upper side of extended data is padded with the mean of
         the first row, the lower side with the mean of the last row, etc.
       - "edge": the data is padded by replicating the edges.
         This is the default mode.
       - "sym": the data is padded by mirroring the data with respect
         to its edges. See numpy.pad().
       - "reflect": the data is padded by reflecting the data with respect
         to its edges, including the edges. See numpy.pad().


    Formulas
    --------
    The radio is divided, in the Fourier domain, by the original
    "Paganin filter" [1]

    $$
    F + 1 + \frac{\delta}{\beta} \lambda D \rho |k|^2   # noqa W605
    $$
    where $k$ is the wave vector, computed as

    $$
    k_l = \frac{1}{P} (\frac{-1}{2} + \frac{l}{N-1})
    $$
    where $P$ is the pixel size, $N$ the number of pixels in one direction,
    and $l \in [0, N-1]$.    # noqa W605
    The factor $\rho$ is either $\pi$ or $1/(4\pi^2)$   # noqa W605
    depending on the convention (default is the former).


    References
    -----------
    [1] D. Paganin Et Al, "Simultaneous phase and amplitude extraction
        from a single defocused image of a homogeneous object",
        Journal of Microscopy, Vol 206, Part 1, 2002
    """

    powers = generate_powers()

    def __init__(
        self,
        shape,
        distance=50,
        energy=20,
        delta_beta=250.0,
        pixel_size=1,
        padding="edge",
        margin=None,
        use_R2C=True,
    ):
        self._init_parameters(
            distance, energy, pixel_size, delta_beta, padding, use_R2C
        )
        self._calc_shape(shape, margin)
        self.compute_filter()

    def _init_parameters(
        self, distance, energy, pixel_size, delta_beta, padding, use_R2C
    ):
        self.distance_cm = distance
        self.distance_micron = distance * 1e4
        self.energy_kev = energy
        self.pixel_size_micron = pixel_size
        self.delta_beta = delta_beta
        self.wavelength_micron = 1.23984199e-3 / self.energy_kev
        self.padding = padding
        self.padding_methods = {
            PaddingMode.ZEROS: self._pad_zeros,
            PaddingMode.MEAN: self._pad_mean,
            PaddingMode.EDGE: self._pad_edge,
            PaddingMode.SYMMETRIC: self._pad_sym,
            PaddingMode.REFLECT: self._pad_reflect,
        }
        self.use_R2C = use_R2C
        if use_R2C:
            self.fft_func = np.fft.rfft2
            self.ifft_func = np.fft.irfft2
        else:
            self.fft_func = np.fft.fft2
            self.ifft_func = np.fft.ifft2

    def _calc_shape(self, shape, margin):
        if np.isscalar(shape):
            shape = (shape, shape)
        else:
            assert len(shape) == 2
        self.shape = shape
        self._set_margin_value(margin)
        self._calc_padded_shape()

    def _set_margin_value(self, margin):
        self.margin = margin
        if margin is None:
            self.shape_inner = self.shape
            self.use_margin = False
            self.margin = ((0, 0), (0, 0))
            return
        self.use_margin = True
        try:
            ((U, D), (L, R)) = margin
        except ValueError:
            raise ValueError("Expected margin in the format ((U, D), (L, R))")
        for val in [U, D, L, R]:
            if type(val) == str and val != "auto":
                raise ValueError("Expected either an integer, or 'auto'")
            if int(val) != val or val < 0:
                raise ValueError("Expected positive integers for margin values")
        self.shape_inner = (self.shape[0] - U - D, self.shape[1] - L - R)

    def _calc_padded_shape(self):
        """
        Compute the padded shape.
        If margin = 0, length_padded = next_power(2*length).
        Otherwise : length_padded = next_power(2*(length - margins))

        Principle
        ----------

        <--------------------- nx_p --------------------->
        |         |        original data       |         |
        < -- Pl - ><-- L -->< -- nx --><-- R --><-- Pr -->
                   <----------- nx0 ----------->

        Pl, Pr : left/right padding length
        L, R : left/right margin
        nx : length of inner data (and length of final result)
        nx0 : length of original data
        nx_p : total length of padded data
        """
        n_y, n_x = self.shape_inner
        n_y_p = self._get_next_power(2 * n_y)
        n_x_p = self._get_next_power(2 * n_x)
        self.shape_padded = (n_y_p, n_x_p)
        self.data_padded = np.zeros((n_y_p, n_x_p), dtype=np.float64)

        ((U, D), (L, R)) = self.margin
        n_y0, n_x0 = self.shape
        self.pad_top_len = (n_y_p - n_y0) // 2
        self.pad_bottom_len = n_y_p - n_y0 - self.pad_top_len
        self.pad_left_len = (n_x_p - n_x0) // 2
        self.pad_right_len = n_x_p - n_x0 - self.pad_left_len

    def _get_next_power(self, n):
        """
        Given a number, get the closest (upper) number p such that
        p is a power of 2, 3, 5 and 7.
        """
        idx = bisect(self.powers, n)
        if self.powers[idx - 1] == n:
            return n
        return self.powers[idx]

    def compute_filter(self):
        nyp, nxp = self.shape_padded
        fftfreq = np.fft.rfftfreq if self.use_R2C else np.fft.fftfreq
        fy = np.fft.fftfreq(nyp, d=self.pixel_size_micron)
        fx = fftfreq(nxp, d=self.pixel_size_micron)
        self._coords_grid = np.add.outer(fy ** 2, fx ** 2)
        #
        k2 = self._coords_grid
        D = self.distance_micron
        L = self.wavelength_micron
        db = self.delta_beta
        self.paganin_filter = 1.0 / (1 + db * L * D * pi * k2)  # HST / savu
        # ~ self.paganin_filter = 1.0 / (1 + db * L * D * k2/ (4*pi))  # Paganin / tomopy

    def pad_with_values(self, data, top_val=0, bottom_val=0, left_val=0, right_val=0):
        """
        Pad the data into `self.padded_data` with values.

        Parameters
        ----------
        data: numpy.ndarray
            data (radio)
        top_val: float or numpy.ndarray, optional
            Value(s) to fill the top of the padded data with.
        bottom_val: float or numpy.ndarray, optional
            Value(s) to fill the bottom of the padded data with.
        left_val: float or numpy.ndarray, optional
            Value(s) to fill the left of the padded data with.
        right_val: float or numpy.ndarray, optional
            Value(s) to fill the right of the padded data with.
        """
        self.data_padded.fill(0)
        Pu, Pd = self.pad_top_len, self.pad_bottom_len
        Pl, Pr = self.pad_left_len, self.pad_right_len
        self.data_padded[:Pu, :] = top_val
        self.data_padded[-Pd:, :] = bottom_val
        self.data_padded[:, :Pl] = left_val
        self.data_padded[:, -Pr:] = right_val
        self.data_padded[Pu:-Pd, Pl:-Pr] = data
        # Transform the data to the FFT layout
        self.data_padded = np.roll(self.data_padded, (-Pu, -Pl), axis=(0, 1))

    def _pad_zeros(self, data):
        return self.pad_with_values(
            data, top_val=0, bottom_val=0, left_val=0, right_val=0
        )

    def _pad_mean(self, data):
        """
        Pad the data at each border with a different constant value.
        The value depends on the padding size:
          - On the left, value = mean(first data column)
          - On the right, value = mean(last data column)
          - On the top, value = mean(first data row)
          - On the bottom, value = mean(last data row)
        """
        return self.pad_with_values(
            data,
            top_val=np.mean(data[0, :]),
            bottom_val=np.mean(data[-1, :]),
            left_val=np.mean(data[:, 0]),
            right_val=np.mean(data[:, -1]),
        )

    def _pad_numpy(self, data, mode):
        data_padded = np.pad(
            data,
            (
                (self.pad_top_len, self.pad_bottom_len),
                (self.pad_left_len, self.pad_right_len),
            ),
            mode=mode.value,
        )
        # Transform the data to the FFT layout
        Pu, Pl = self.pad_top_len, self.pad_left_len
        return np.roll(data_padded, (-Pu, -Pl), axis=(0, 1))

    def _pad_edge(self, data):
        self.data_padded = self._pad_numpy(data, mode=PaddingMode.EDGE)

    def _pad_sym(self, data):
        self.data_padded = self._pad_numpy(data, mode=PaddingMode.SYMMETRIC)

    def _pad_reflect(self, data):
        self.data_padded = self._pad_numpy(data, mode=PaddingMode.REFLECT)

    def pad_data(self, data, padding_method=None):
        padding_method = padding_method or self.padding
        padding_method = PaddingMode.from_value(padding_method)
        if padding_method not in self.padding_methods:
            raise ValueError(
                "Unknown padding method %s. Available are: %s"
                % (padding_method, str(list(self.padding_methods.keys())))
            )
        pad_func = self.padding_methods[padding_method]
        pad_func(data)
        return self.data_padded

    def apply_filter(self, radio, padding_method=None):
        self.pad_data(radio, padding_method=padding_method)
        radio_f = self.fft_func(self.data_padded)
        radio_f *= self.paganin_filter
        radio_filtered = self.ifft_func(radio_f).real
        s0, s1 = self.shape_inner
        ((U, _), (L, _)) = self.margin
        return radio_filtered[U : U + s0, L : L + s1]

    def lmicron_to_db(self, Lmicron):
        """
        Utility to convert the "Lmicron" parameter of PyHST
        to a value of delta/beta.
        Please see the doc of nabu.preproc.phase.lmicron_to_db()
        """
        return lmicron_to_db(Lmicron, self.energy_kev, self.distance_micron)

    __call__ = apply_filter



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "12/08/2021"

import logging
import os
import shutil
import tempfile
from silx.gui import qt
from orangecontrib.tomwer.widgets.control.FilterOW import NameFilterOW
from silx.gui.utils.testutils import TestCaseQt
from tomwer.core.utils.scanutils import MockHDF5, MockEDF
from tomwer.core.scan.edfscan import EDFTomoScan
from silx.gui.utils.testutils import SignalListener


logging.disable(logging.INFO)


class NameFilterOWCI(NameFilterOW):
    sigScanReady = qt.Signal(str)
    """signal emitted when a scan is ready; Parameter is folder path of the
    scan"""

    def _signalScanReady(self, scan):
        self.sigScanReady.emit(scan.path)


class TestFilterWidget(TestCaseQt):
    """class testing the DarkRefWidget"""

    def setUp(self):
        TestCaseQt.setUp(self)
        self.widget = NameFilterOWCI(parent=None)

        self.tmpdir = tempfile.mkdtemp()
        # create hdf5 dataset
        self._hdf5with_mytotodir = MockHDF5(
            scan_path=os.path.join(self.tmpdir, "mytotodir120"), n_proj=10
        ).scan
        self._hdf5without_mytotodir = MockHDF5(
            scan_path=os.path.join(self.tmpdir, "totodir120"), n_proj=10
        ).scan
        # create edf dataset
        mockEDF1 = MockEDF(
            scan_path=os.path.join(self.tmpdir, "mytotodir20"), n_radio=10
        )
        self.edfwith_mytotodir = EDFTomoScan(scan=mockEDF1.scan_path)

        mockEDF2 = MockEDF(scan_path=os.path.join(self.tmpdir, "dir120"), n_radio=10)
        self.edfwithout_mytotodir = EDFTomoScan(scan=mockEDF2.scan_path)

        # add a signal listener
        self.signalListener = SignalListener()
        self.widget.sigScanReady.connect(self.signalListener)

        # define the unix file pattern
        self.widget.setPattern("*mytotodir*")

    def tearDown(self):
        shutil.rmtree(self.tmpdir)
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        TestCaseQt.tearDown(self)

    def testEDF(self):
        """Make sure filtering works with EDF"""
        self.assertEqual(self.signalListener.callCount(), 0)
        self.widget.applyfilter(self.edfwith_mytotodir)
        self.qapp.processEvents()
        self.assertEqual(self.signalListener.callCount(), 1)
        self.widget.applyfilter(self.edfwithout_mytotodir)
        self.qapp.processEvents()
        self.assertEqual(self.signalListener.callCount(), 1)

    def testHDF5(self):
        """Make sure filtering works with HDF5"""
        self.assertEqual(self.signalListener.callCount(), 0)
        self.widget.applyfilter(self._hdf5without_mytotodir)
        self.qapp.processEvents()
        self.assertEqual(self.signalListener.callCount(), 0)
        self.widget.applyfilter(self._hdf5with_mytotodir)
        self.qapp.processEvents()
        self.assertEqual(self.signalListener.callCount(), 1)



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "19/10/2021"


from tomwer.gui.reconstruction.nabu.slurm import SlurmSettingsWidget
from tomwer.core.settings import SlurmSettings
from silx.gui.utils.testutils import TestCaseQt
from silx.gui import qt
from tomwer.test.utils import skip_gui_test
import pytest


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestSlurmWidget(TestCaseQt):
    def setUp(self):
        TestCaseQt.setUp(self)
        self.slurmWidget = SlurmSettingsWidget(parent=None)

    def tearDown(self):
        self.slurmWidget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.slurmWidget.close()
        self.slurmWidget = None
        TestCaseQt.tearDown(self)

    def testGetConfiguration(self):
        self.slurmWidget._slurmCB.setChecked(False)
        dict_res = self.slurmWidget.getConfiguration()
        expected_dict = {
            "active_slurm": False,
            "cores": SlurmSettings.N_CORES_PER_WORKER,
            "n_workers": SlurmSettings.N_WORKERS,
            "memory": SlurmSettings.MEMORY_PER_WORKER,
            "queue": SlurmSettings.QUEUE,
            "n_gpus": SlurmSettings.N_GPUS_PER_WORKER,
        }
        assert dict_res == expected_dict, f"{dict_res} vs {expected_dict}"

    def testSetConfiguration(self):
        self.slurmWidget.setConfiguration(
            {
                "cores": 2,
                "n_workers": 3,
                "memory": 156,
                "queue": "test-queue",
                "n_gpus": 5,
            }
        )

        assert self.slurmWidget.getNCores() == 2
        assert self.slurmWidget.getNWorkers() == 3
        assert self.slurmWidget.getMemory() == 156
        assert self.slurmWidget.getQueue() == "test-queue"
        assert self.slurmWidget.getNGPU() == 5



from silx.gui import qt
from tomwer.core.utils.device import CudaDevice
from tomwer.core.utils.device import _CudaPlatformBase
import logging

_logger = logging.getLogger(__name__)


class _PlatformSelectorBase:
    """simple interface for platform selection"""

    def getDevices(self):
        raise NotImplementedError("Base class")

    def getSelectedDevide(self):
        raise NotImplementedError("Base class")


class CudaPlatfornComboBox(qt.QWidget, _CudaPlatformBase):
    NONE_FOUND = "no cuda compliant gpu found"

    def __init__(self, parent=None):
        super(CudaPlatfornComboBox, self).__init__(parent=parent)
        self.setLayout(qt.QHBoxLayout())

        label = qt.QLabel("GPU Device:", self)
        self.layout().addWidget(label)
        self.deviceSelector = qt.QComboBox(self)
        self.deviceSelector.addItem(CudaDevice.NONE_FOUND)
        self.layout().addWidget(self.deviceSelector)

        devices = self.getExistingDevices()
        if len(devices) > 0:
            self.deviceSelector.clear()
            for device in devices:
                self.deviceSelector.addItem(device.name)

    def getSelectedDevice(self):
        """Return platform id and device id"""
        if self.deviceSelector.currentText() == self.NONE_FOUND:
            return None
        else:
            return CudaDevice(self.deviceSelector.currentText())


class CudaPlatfornGroup(qt.QGroupBox, _CudaPlatformBase):
    """
    GroupBox listing all the existing cuda device and allow user to select the
    one to use for reconstruction
    """

    selectionChanged = qt.Signal(list)

    def __init__(self, parent=None):
        qt.QGroupBox.__init__(self, parent=parent)
        _CudaPlatformBase.__init__(self)
        self.setLayout(qt.QVBoxLayout())
        self.setTitle("GPU devices to use")
        self._devices = {}
        try:
            existing_devices = self.getExistingDevices()
        except Exception as e:
            _logger.error(e)
        else:
            for device in existing_devices:
                qcb = qt.QCheckBox(device.name)
                self.layout().addWidget(qcb)
                self._devices[device] = qcb
                qcb.stateChanged.connect(self._selectionUpdated)

    def activate_all(self):
        for device in self._devices:
            self._devices[device].setChecked(True)

    def getSelectedDevices(self):
        """

        :return: list of "CudaDevice" selected by the user
        :rtype:list
        """
        selected = []
        for device in self._devices:
            if self._devices[device].isChecked():
                selected.append(device)
            return selected

    def removeSelection(self):
        for device in self._devices:
            self._devices[device].setChecked(False)

    def setDevices(self, selection):
        """
        set selected devices.

        :param list selection: selected devices, list of "CudaDevice" or int
        """
        if selection is None:
            self.activate_all()
            return

        def device_to_index():
            res = {}
            for device, checkbox in self._devices.items():
                res[device._id] = checkbox
            return res

        indexes_to_cb = device_to_index()
        for device in selection:
            if isinstance(selection, CudaDevice):
                id_ = device.id_
            else:
                id_ = device

            if id_ in indexes_to_cb:
                indexes_to_cb[id_].setChecked(True)

        self._selectionUpdated()

    def _selectionUpdated(self, *args, **kwargs):
        self.selectionChanged.emit(self.getSelectedDevices())



# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/


__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "01/10/2018"


from queue import Queue
from silx.gui import qt
import logging
from tomwer.core.scan.scanbase import TomwerScanBase
from processview.core.superviseprocess import SuperviseProcess
import copy
from tomwer.core.process.reconstruction.lamino.tofu import LaminoReconstruction
from tomwer.core.utils.Singleton import singleton
from tomwer.web.client import OWClient

logger = logging.getLogger(__name__)


@singleton
class LaminoReconstructionStack(qt.QObject, Queue, OWClient):
    """
    Manage a stack of lamino (tofu) reconstruction
    """

    sigReconsFinished = qt.Signal(TomwerScanBase)
    """Signal emitted when a new reconstruction is ended"""
    sigReconsFailed = qt.Signal(TomwerScanBase)
    """Signal emitted when a new reconstruction has failed"""
    sigReconsMissParams = qt.Signal(TomwerScanBase)
    """Signal emitted when some paramters are missing for the reconstruction"""
    sigReconsStarted = qt.Signal(TomwerScanBase)
    """Signal emitted when a new reconstruction is started"""

    def __init__(self, process_id=None):
        qt.QObject.__init__(self)
        Queue.__init__(self)
        OWClient.__init__(self)
        self.process_id = process_id
        self.reconsThread = _LaminoReconsThread(process_id=process_id)
        self.reconsThread.sigThReconsFinished.connect(self._dealWithFinishedRecons)
        self.reconsThread.sigThReconsFailed.connect(self._dealWithFailedRecons)
        self.reconsThread.sigThMissingParams.connect(self._dealWithThMissingParams)
        self._forceSync = False

    def add(
        self,
        recons_obj,
        recons_params,
        additional_opts,
        scan_id,
        remove_existing,
        callback,
    ):
        """
        add a reconstruction and will run it as soon as possible

        :param recons_obj: reconstructor, keeping trace of preprocess flat field
                           correction for example.
        :type recons_obj: LaminoReconstruction
        :param dict reconsParams: parameters of the reconstruction
        :param dict additional_opts: not managed directly by the gui
        :param scan_id: the folder of the acquisition to reconstruct
        :type: TomoBase
        :param bool remove_existing: if True then remove output dir before
                                     reconstruction
        :param callback: function to call after the reconstruction execution
        """
        assert isinstance(recons_obj, LaminoReconstruction)
        Queue.put(
            self,
            (
                recons_obj,
                recons_params,
                additional_opts,
                scan_id,
                remove_existing,
                callback,
            ),
        )
        if self.canExecNext():
            self.execNext()

    def execNext(self):
        """
        Launch the next reconstruction if any
        """
        if Queue.empty(self):
            return

        assert not self.reconsThread.isRunning()
        (
            recons_obj,
            recons_params,
            additional_opts,
            scan,
            remove_existing,
            callback,
        ) = Queue.get(self)
        self.sigReconsStarted.emit(scan)
        info = "start lamino reconstuction of {}".format(str(scan))
        logger.processStarted(info)
        self.reconsThread.init(
            recons_obj=recons_obj,
            recons_params=recons_params,
            additional_opts=additional_opts,
            scan_id=scan,
            remove_existing=remove_existing,
        )
        self.reconsThread.sigThReconsFinished.connect(callback)
        self.reconsThread.start()
        if self._forceSync is True:
            self.reconsThread.wait()

    def canExecNext(self):
        """
        Can we launch an ftserie reconstruction.
        Reconstruction can't be runned in parallel

        :return: True if no reconstruction is actually running
        """
        return not self.reconsThread.isRunning()

    def _dealWithFinishedRecons(self, scan):
        assert isinstance(scan, TomwerScanBase)
        info = "lamino reconstruction of {} finished".format(str(scan))
        logger.processSucceed(info)
        self.sigReconsFinished.emit(scan)
        self.execNext()

    def _dealWithThMissingParams(self, scan):
        assert isinstance(scan, TomwerScanBase)
        self.sigReconsMissParams.emit(scan)
        self.execNext()

    def _dealWithFailedRecons(self, scan):
        assert isinstance(scan, TomwerScanBase)
        info = "reconstruction of {} failed".format(str(scan))
        logger.processFailed(info)
        self.sigReconsFailed.emit(scan)
        self.execNext()

    def setMockMode(self, b):
        self.reconsThread.setMockMode(b)
        self.execNext()

    def setForceSync(self, b=True):
        self._forceSync = b


class _LaminoReconsThread(qt.QThread, SuperviseProcess):
    """Thread used for running lamino reconstrucion using Tofu"""

    sigThReconsFinished = qt.Signal(TomwerScanBase)
    "Emitted if reconstruction ended with success"
    sigThReconsFailed = qt.Signal(TomwerScanBase)
    "Emitted if reconstruction failed"
    sigThMissingParams = qt.Signal(TomwerScanBase)
    "Emitted if missing some reconstruction parameters"

    def __init__(self, process_id):
        try:
            qt.QThread.__init__(self, process_id=process_id)
        except TypeError:
            qt.QThread.__init__(self)
        SuperviseProcess.__init__(self, process_id=process_id)
        self.scan = None
        self.recons_params = None
        self.additional_opts = None
        self.callback = None
        self.remove_existing = False

    def init(
        self, recons_obj, recons_params, additional_opts, remove_existing, scan_id
    ):
        assert isinstance(recons_obj, LaminoReconstruction)
        self.recons_obj = recons_obj
        self.recons_params = recons_params
        self.additional_opts = additional_opts
        self.remove_existing = remove_existing
        self.scan = scan_id

    def run(self):
        # we need to create an other LaminoReconstruction object. Because
        # recons parameters might have changed bu we want to be up to date
        # regarding the flat field preprocessed already if any to avoid
        # several computation
        recons_obj = copy.deepcopy(self.recons_obj)
        recons_obj.reconstruction_parameters = self.recons_params
        recons_obj.additional_reco_options = self.additional_opts
        try:
            recons_obj.process(scan=self.scan)
        except ValueError as error:
            logger.warning(error)
            self.sigThMissingParams.emit(
                "Some parameters are missing or are incoherent"
            )
        except Exception as error:
            logger.error(
                "fail to run lamino reconstruction for %s reason is "
                "%s" % (self.scan, error)
            )
            self.sigThReconsFailed.emit(
                "fail to run reconstruction %s" % self.scan.path
            )
        else:
            self.sigThReconsFinished.emit(self.scan)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "30/10/2018"

"""Include default settings for the tofu widget"""

# default parameters for slice stack
SLICE_STACK_RANGE_HS = 50
"""default range half size for slice stack reconstruction"""
SLICE_STACK_STEP_SIZE = 1
"""default step size for slice stack reconstruction"""
SLICE_STACK_NB_SLICES = 100
"""default number of slices for slice stack reconstruction"""

# default parameters for rotation center
ROT_CENTER_RANGE_HS = 25
"""default range half size for rotation center reconstruction"""
ROT_CENTER_STEP_SIZE = 0.5
"""default step size for rotation center reconstruction"""
ROT_CENTER_NB_SLICES = 50
"""default number of slices for rotation center reconstruction"""

# default parameters for lamino angle
LAMINO_ANG_RANGE_HS = 0.5
"""default range half size for the lamino angle reconstruction"""
LAMINO_ANG_STEP_SIZE = 0.01
"""default step size for the lamino angle reconstruction"""
LAMINO_ANG_NB_SLICES = 100
"""default number of slices for lamino angle reconstruction"""

# default parameters for psi angle
PSI_ANG_RANGE_HS = 0.5
"""default range half size for the psi angle reconstruction"""
PSI_ANG_STEP_SIZE = 0.01
"""default step size for the psi angle reconstruction"""
PSI_ANG_NB_SLICES = 100
"""default number of slices for psi angle reconstruction"""


_N_SLICE_LIMITS_HIGH = 350
"""If the number of slices is upper than this limits then this will generate a
warning and display the information into"""

_COLOR_HIGH = 250, 0, 0

_N_SLICE_LIMITS_MEDIUM = 100
"""If the number of slices is upper this than limits then this will display the
cut selection information into"""

_COLOR_MEDIUM = 200, 150, 0



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

"""
This module is used to define the process of the reference creator.
This is related to the issue #184
"""

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "19/07/2018"

import fnmatch
import logging
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
import tomwer.version
import os
import re

from tomwer.core.process.task import Task

_logger = logging.getLogger(__name__)


class BaseFilter(object):
    """
    Apply a filter to an object
    """

    def description(self):
        pass

    def isFiltered(self, value):
        """
        Return True if the value not filtered

        """
        raise NotImplementedError("Base class")


class _PatternBaseFilter(BaseFilter):
    """Filter based on a pattern"""

    def __init__(self, pattern):
        BaseFilter.__init__(self)
        self.setPattern(pattern)

    def setPattern(self, pattern):
        """
        compile th filter for the given pattern
        :param str pattern:
        """
        self._pattern = pattern

    def getPattern(self):
        return self._pattern


class RegularExpressionFilter(_PatternBaseFilter):
    """Filter a string based on a defined pattern"""

    def __init__(self, pattern):
        _PatternBaseFilter.__init__(self, pattern=pattern)

    def setPattern(self, pattern):
        """
        compile th filter for the given pattern
        :param str pattern:
        """
        super().setPattern(pattern)
        if self._pattern is not None:
            try:
                self._filter = re.compile(self._pattern)
            except re.error as e:
                self.unvalidPatternDefinition(self, self._pattern, e)
                _logger.error(e)

    def description(self):
        return "Filter a string base on a regular expression"

    def isFiltered(self, value):
        try:
            match = self._filter.match(value) is None
        except Exception:
            return False
        else:
            return match

    def unvalidPatternDefinition(self, cls, pattern, error):
        _logger.error(f"{cls} {pattern} is not a valid pattern. Error is {error}")


class UnixFileNamePatternFilter(_PatternBaseFilter):
    """Filter a string based on 'fnmatch' module (unix filename pattern
    matching)"""

    def __init__(self, pattern):
        _PatternBaseFilter.__init__(self, pattern)

    def description(self):
        return "Filter a string base on a glob (unix style pathname)"

    def isFiltered(self, value):
        try:
            match = fnmatch.fnmatch(value, self._pattern)
        except Exception:
            match = False
        return not match


class FileNameFilter(
    _PatternBaseFilter, Task, input_names=("data",), output_names=("data",)
):
    """Filter than can call several filter type"""

    UNIX_PATTERN_FILTER = "unix file name pattern"

    REGULAR_EXP_FILTER = "regular expression"

    FILTER_TYPES = UNIX_PATTERN_FILTER, REGULAR_EXP_FILTER

    _DEFAULT_FILTER_TYPE = FILTER_TYPES[0]

    _FILTER_CONSTR = {
        UNIX_PATTERN_FILTER: UnixFileNamePatternFilter,
        REGULAR_EXP_FILTER: RegularExpressionFilter,
    }

    def __init__(self, varinfo=None, inputs=None, label=None):
        Task.__init__(self, varinfo=varinfo, inputs=inputs, label=label)
        pattern = inputs.get("pattern", None)
        if pattern is None:
            raise ValueError("'pattern' should be provided")
        filter_type = inputs.get("filter_type", self._DEFAULT_FILTER_TYPE)
        self._invert = inputs.get("invert_result", False)

        self._filter = self._FILTER_CONSTR[filter_type](pattern)

        _PatternBaseFilter.__init__(self, pattern)

    @property
    def invert_result(self):
        return self._invert

    @property
    def filter(self):
        return self._filter

    def isFiltered(self, value):
        result = self._filter.isFiltered(value)
        if self.invert_result:
            return not result
        else:
            return result

    def run(self):
        scan = self.inputs.data
        if scan is None:
            raise ValueError("'data' should be provided")
        if isinstance(scan, dict):
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        if scan is None:
            return False
        assert isinstance(scan, TomwerScanBase)

        if not self.isFiltered(os.path.basename(scan.path)):
            if self._return_dict:
                self.outputs.data = scan.to_dict()
            else:
                self.outputs.data = scan
        else:
            self.outputs.data = None

    @staticmethod
    def program_name():
        return "scan filter"

    @staticmethod
    def program_version():
        """version of the program used for this processing"""
        return tomwer.version.version

    @staticmethod
    def definition():
        """definition of the process"""
        return "filter a scan according to his name"



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "30/07/2020"


from tomwer.core.process.task import Task
from tomwer.core.scan.blissscan import BlissScan
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from nxtomomill import converter as nxtomomill_converter

try:
    # new HDF5Config class
    from nxtomomill.io.config import TomoHDF5Config as HDF5Config
except ImportError:
    from nxtomomill.io.config import HDF5Config
import os
import logging

_logger = logging.getLogger(__name__)


class NxTomomillProcess(Task, input_names=("data",), output_names=("data",)):
    """
    Process to convert from a bliss dataset to a nexus compliant dataset
    """

    @staticmethod
    def deduce_output_file_path(master_file_name, entry, outputdir=None):
        if outputdir is not None:
            file_dir = outputdir
        else:
            file_dir = os.path.dirname(master_file_name)
        file_name = os.path.basename(master_file_name)
        if "." in file_name:
            file_name = "".join(file_name.split(".")[:-1])

        entry_for_file_name = entry.lstrip("/")
        entry_for_file_name = entry_for_file_name.replace("/", "_")
        entry_for_file_name = entry_for_file_name.replace(".", "_")
        entry_for_file_name = entry_for_file_name.replace(":", "_")
        output_file_name = "_".join(
            (os.path.splitext(file_name)[0], entry_for_file_name + ".nx")
        )
        return os.path.join(file_dir, output_file_name)

    def run(self):
        scan = self.inputs.data
        if scan is None:
            self.outputs.data = None
            return

        _logger.processStarted("Start translate {} to NXTomo".format(str(scan)))
        if isinstance(scan, dict):
            scan = BlissScan.from_dict(scan)

        if not isinstance(scan, BlissScan):
            raise TypeError("Scan is expected to be an instance of BlissScan")
        output_file_path = self.deduce_output_file_path(
            master_file_name=scan.master_file, entry=scan.entry
        )
        _logger.info(" ".join(("write", str(scan), "to", output_file_path)))
        try:
            configuration = HDF5Config.from_dict(self.get_configuration())
            configuration.input_file = scan.master_file
            configuration.output_file = output_file_path
            configuration.entries = (scan.entry,)
            # overwrite some parameters
            configuration.single_file = False
            configuration.overwrite = True
            configuration.file_extension = ".nx"
            convs = nxtomomill_converter.from_h5_to_nx(
                configuration=configuration,
            )
        except Exception as e:
            _logger.processFailed(
                "Fail to convert from bliss file: %s to NXTomo."
                "Conversion error is: %s" % (scan.master_file, e)
            )
            self.outputs.data = None
            return
        else:
            for conv in convs:
                conv_file, conv_entry = conv
                scan_converted = HDF5TomoScan(scan=conv_file, entry=conv_entry)
                _logger.processSucceed(
                    "{} has been translated to {}"
                    "".format(str(scan), str(scan_converted))
                )
                self.outputs.data = scan_converted

    def set_configuration(self, configuration):
        # for now the NXProcess cannot be tune
        if isinstance(configuration, HDF5Config):
            self._settings = configuration.to_dict()
        elif isinstance(configuration, dict):
            self._settings = configuration
        else:
            raise TypeError("invalid type: {}".format(type(configuration)))
